// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "drawstuff.h"
#include "surface.h"
#include <iostream>

TEXTURE_HANDLE white_texture;

static bool first_run = true;

void InitDrawStuff(IGraphicsSystem* IGS)
{
    if (first_run == true)
    {
        first_run = false;

        Surface texture(32, 32, R8G8B8);
        for (int i = 0; i < 32 * 32 * 3; i++)
        {
            texture.buffer[i] = 255;
        }

        white_texture = IGS->CreateTexture(texture, false);
    }
}

void ShutdownDrawStuff(IGraphicsSystem* IGS)
{
    IGS->DeleteTexture(white_texture);
}

void DrawBox(IGraphicsSystem* IGS, const float sides[3], matrix44 transformation_matrix)
{
    InitDrawStuff(IGS);

    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, transformation_matrix);
    IGS->SetTexture(0, white_texture);
    IGS->SetCullMode(CULL_CW);

    float lx = sides[0]*0.5f;
    float ly = sides[1]*0.5f;
    float lz = sides[2]*0.5f;

    vector3 cube_corners[8];

    cube_corners[0] = vector3(-lx,  ly,  lz);
    cube_corners[1] = vector3(-lx,  ly, -lz);
    cube_corners[2] = vector3( lx,  ly, -lz);
    cube_corners[3] = vector3( lx,  ly,  lz);
    cube_corners[4] = vector3(-lx, -ly,  lz);
    cube_corners[5] = vector3(-lx, -ly, -lz);
    cube_corners[6] = vector3( lx, -ly, -lz);
    cube_corners[7] = vector3( lx, -ly,  lz);

    VERTEX_XYZ_NORMAL points[4];

    unsigned short indices[] = { 0, 1, 2, 2, 3, 0};

    points[0].point = cube_corners[0];
    points[1].point = cube_corners[1];
    points[2].point = cube_corners[5];
    points[3].point = cube_corners[4];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(-1.0f, 0.0f, 0.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL,(float*)points, 4, indices, 6);

    points[0].point = cube_corners[1];
    points[1].point = cube_corners[2];
    points[2].point = cube_corners[6];
    points[3].point = cube_corners[5];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(0.0f, 0.0f, -1.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL,(float*)points, 4, indices, 6);

    points[0].point = cube_corners[2];
    points[1].point = cube_corners[3];
    points[2].point = cube_corners[7];
    points[3].point = cube_corners[6];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(1.0f, 0.0f, 0.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL,(float*)points, 4, indices, 6);

    points[0].point = cube_corners[3];
    points[1].point = cube_corners[0];
    points[2].point = cube_corners[4];
    points[3].point = cube_corners[7];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(0.0f, 0.0f, 1.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL,(float*)points, 4, indices, 6);

    points[0].point = cube_corners[0];
    points[1].point = cube_corners[3];
    points[2].point = cube_corners[2];
    points[3].point = cube_corners[1];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(0.0f, 1.0f, 0.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL,(float*)points, 4, indices, 6);

    points[0].point = cube_corners[4];
    points[1].point = cube_corners[5];
    points[2].point = cube_corners[6];
    points[3].point = cube_corners[7];
    points[0].normal = points[1].normal = points[2].normal = points[3].normal = vector3(0.0f, -1.0f, 0.0f);

    IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL, (float*)points, 4, indices, 6);

}

// This is recursively subdivides a triangular area (vertices p1,p2,p3) into
// smaller triangles, and then draws the triangles. All triangle vertices are
// scaled to a distance of radius from the origin (p1,p2,p3 are assumed
// to be already scaled).

void drawPatch (IGraphicsSystem* IGS, float p1[3], float p2[3], float p3[3], int level, float radius)
{
  int i;
  if (level > 0) {
    float q1[3],q2[3],q3[3];		 // sub-vertices
    for (i=0; i<3; i++) {
      q1[i] = 0.5f*(p1[i]+p2[i]);
      q2[i] = 0.5f*(p2[i]+p3[i]);
      q3[i] = 0.5f*(p3[i]+p1[i]);
    }
    float length1 = (float)(radius/sqrt(q1[0]*q1[0]+q1[1]*q1[1]+q1[2]*q1[2]));
    float length2 = (float)(radius/sqrt(q2[0]*q2[0]+q2[1]*q2[1]+q2[2]*q2[2]));
    float length3 = (float)(radius/sqrt(q3[0]*q3[0]+q3[1]*q3[1]+q3[2]*q3[2]));
    for (i=0; i<3; i++) {
      q1[i] *= length1;
      q2[i] *= length2;
      q3[i] *= length3;
    }
    drawPatch (IGS, p1, q1, q3, level-1, radius);
    drawPatch (IGS, q1, p2, q2, level-1, radius);
    drawPatch (IGS, q1, q2, q3, level-1, radius);
    drawPatch (IGS, q3, q2, p3, level-1, radius);
  }
  else {

    VERTEX_XYZ_NORMAL points[3];

    points[0].point = vector3(p1[0], p1[1], p1[2]);
    points[0].normal= vector3(p1[0], p1[1], p1[2]);
    points[1].point = vector3(p2[0], p2[1], p2[2]);
    points[1].normal= vector3(p2[0], p2[1], p2[2]);
    points[2].point = vector3(p3[0], p3[1], p3[2]);
    points[2].normal= vector3(p3[0], p3[1], p3[2]);

    IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL, (float*)points, 3);
  }
}

// draw a sphere of radius 1

static int sphere_quality = 1;

void drawSphere(IGraphicsSystem* IGS, float radius)
{
  // icosahedron data for an icosahedron of radius 1.0
# define ICX 0.525731112119133606f
# define ICZ 0.850650808352039932f
  float idata[12][3] = {
    {-ICX * radius, 0 * radius, ICZ * radius},
    {ICX * radius, 0 * radius, ICZ * radius},
    {-ICX * radius, 0 * radius, -ICZ * radius},
    {ICX * radius, 0 * radius, -ICZ * radius},
    {0 * radius, ICZ * radius, ICX * radius},
    {0 * radius, ICZ * radius, -ICX * radius},
    {0 * radius, -ICZ * radius, ICX * radius},
    {0 * radius, -ICZ * radius, -ICX * radius},
    {ICZ * radius, ICX * radius, 0 * radius},
    {-ICZ * radius, ICX * radius, 0 * radius},
    {ICZ * radius, -ICX * radius, 0 * radius},
    {-ICZ * radius, -ICX * radius, 0 * radius}
  };

  int index[20][3] = {
    {0, 4, 1},	  {0, 9, 4},
    {9, 5, 4},	  {4, 5, 8},
    {4, 8, 1},	  {8, 10, 1},
    {8, 3, 10},   {5, 3, 8},
    {5, 2, 3},	  {2, 7, 3},
    {7, 10, 3},   {7, 6, 10},
    {7, 11, 6},   {11, 0, 6},
    {0, 1, 6},	  {6, 1, 10},
    {9, 0, 11},   {9, 11, 2},
    {9, 2, 5},	  {7, 2, 11},
  };

    int i;
    for (i=0; i<20; i++) {
      drawPatch (IGS, &idata[index[i][2]][0],&idata[index[i][1]][0],
        &idata[index[i][0]][0],sphere_quality, radius);
    }
}

void DrawSphere(IGraphicsSystem* IGS, const float radius, matrix44 transformation_matrix)
{
    InitDrawStuff(IGS);

    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, transformation_matrix);
    IGS->SetTexture(0, white_texture);
    IGS->SetCullMode(CULL_CW);

    drawSphere(IGS, radius);
}


// draw a capped cylinder of length l and radius r, aligned along the x axis

static int capped_cylinder_quality = 3;

static void drawCappedCylinder (IGraphicsSystem* IGS, float l, float r)
{
  int i,j;
  float tmp,nx,ny,nz,start_nx,start_ny,a,ca,sa;
  // number of sides to the cylinder (divisible by 4):
  const int n = capped_cylinder_quality*4;

  VERTEX_XYZ_NORMAL points[4];

  unsigned short indices[] = { 0, 2, 1, 1, 2, 3};

  l *= 0.5;
  a = float(3.1428*2.0)/float(n);
  sa = (float) sin(a);
  ca = (float) cos(a);

  // draw cylinder body
  ny=1; nz=0;		  // normal vector = (0,ny,nz)
  for (i=0; i<=n; i++) {
    points[2].point = points[0].point; points[2].normal = points[0].normal;
    points[3].point = points[1].point; points[3].normal = points[1].normal;

    points[0].point = vector3(ny*r,nz*r,l);
    points[0].normal = vector3(ny,nz,0);
    points[1].point = vector3(ny*r,nz*r,-l);
    points[1].normal = vector3(ny,nz,0);

    if (i > 0)
    {
        IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL, (float*)points, 4, indices, 6);
    }

    // rotate ny,nz
    tmp = ca*ny - sa*nz;
    nz = sa*ny + ca*nz;
    ny = tmp;
  }


  // draw first cylinder cap
  start_nx = 0;
  start_ny = 1;
  for (j=0; j<(n/4); j++) {
    // get start_n2 = rotated start_n
    float start_nx2 =  ca*start_nx + sa*start_ny;
    float start_ny2 = -sa*start_nx + ca*start_ny;
    // get n=start_n and n2=start_n2
    nx = start_nx; ny = start_ny; nz = 0;
    float nx2 = start_nx2, ny2 = start_ny2, nz2 = 0;
    for (i=0; i<=n; i++) {
        points[2].point = points[0].point; points[2].normal = points[0].normal;
        points[3].point = points[1].point; points[3].normal = points[1].normal;

    points[0].point = vector3(ny2*r,nz2*r,l+nx2*r);
    points[0].normal = vector3(ny2,nz2,nx2);
    points[1].point = vector3(ny*r,nz*r,l+nx*r);
    points[1].normal = vector3(ny,nz,nx);

    if (i > 0)
    {
        IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL, (float*)points, 4, indices, 6);
    }

      // rotate n,n2
      tmp = ca*ny - sa*nz;
      nz = sa*ny + ca*nz;
      ny = tmp;
      tmp = ca*ny2- sa*nz2;
      nz2 = sa*ny2 + ca*nz2;
      ny2 = tmp;
    }
    start_nx = start_nx2;
    start_ny = start_ny2;
  }

  // draw second cylinder cap
  start_nx = 0;
  start_ny = 1;
  for (j=0; j<(n/4); j++) {
    // get start_n2 = rotated start_n
    float start_nx2 = ca*start_nx - sa*start_ny;
    float start_ny2 = sa*start_nx + ca*start_ny;
    // get n=start_n and n2=start_n2
    nx = start_nx; ny = start_ny; nz = 0;
    float nx2 = start_nx2, ny2 = start_ny2, nz2 = 0;
    for (i=0; i<=n; i++) {
        points[2].point = points[0].point; points[2].normal = points[0].normal;
        points[3].point = points[1].point; points[3].normal = points[1].normal;

    points[0].point = vector3(ny*r,nz*r,-l+nx*r);
    points[0].normal = vector3(ny,nz,nx);
    points[1].point = vector3(ny2*r,nz2*r,-l+nx2*r);
    points[1].normal = vector3(ny2,nz2,nx2);

    if (i > 0)
    {
        IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_NORMAL, (float*)points, 4, indices, 6);
    }

      // rotate n,n2
      tmp = ca*ny - sa*nz;
      nz = sa*ny + ca*nz;
      ny = tmp;
      tmp = ca*ny2- sa*nz2;
      nz2 = sa*ny2 + ca*nz2;
      ny2 = tmp;
    }
    start_nx = start_nx2;
    start_ny = start_ny2;
  }
}

void DrawCappedCylinder(IGraphicsSystem* IGS, const float length, const float radius, matrix44 transformation_matrix)
{
    InitDrawStuff(IGS);

    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, transformation_matrix);
    IGS->SetTexture(0, white_texture);
    IGS->SetCullMode(CULL_CW);

    drawCappedCylinder(IGS, length, radius);
}

