// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include "editor_main.h"
#include "editor_pick.h"
#include "mouse.h"
#include "keyboard.h"
#include "filenames.h"
#include "hud_renderer.h"
#include "object_renderer.h"
#include "gettime.h"

Editor::Editor(IGraphicsSystem* IGS, Level* level)
{
    // Store the pointer to the rendering interface.
    this->IGS = IGS;

    // Create a white texture
    Surface texture(32, 32, R8G8B8);
    for (int i = 0; i < 32 * 32 * 3; i++)
    {
        texture.buffer[i] = 255;
    }

    white_texture = IGS->CreateTexture(texture, true);

    camera = vector3(200.0f, 100.0f, 300.0f);
    direction = vector3(0.0f, 1.0f, 1.0f).normalize();
    zoom = 100.0f;
    lookat = camera - zoom * direction;
    up = vector3(0.0f, 1.0f, 0.0f);
    angle_y = 0.0f;
    angle_x = 3.14f / 180.0f * 0.0f;

    editor_world = new EditorWorld(level, IGS);

    horizontal_plane = 30.0f;
}

Editor::~Editor()
{
    // Delete the textures
    IGS->DeleteTexture(white_texture);

    delete editor_world;
}

void Editor::Render(matrix44 perspective_matrix, matrix44 view_matrix)
{
    static long start = GetTime();
    long elapsed_time = GetTime() - start;  
    float movement = (float)elapsed_time / 15.0f;
    start = GetTime();

    if (Keyboard::Instance().KeyPressed(SDLK_LSHIFT) || Keyboard::Instance().KeyPressed(SDLK_RSHIFT))
    {
        movement *= 5.0f;
    }

    if (Keyboard::Instance().KeyPressed(SDLK_UP))
    {
        lookat += RotateRadMatrix44('y', angle_y) 
        * vector3(0, 0, -1.0f) * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_DOWN))
    {
        lookat += RotateRadMatrix44('y', angle_y) 
        * vector3(0, 0, 1.0f) * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_LEFT))
    {
        lookat += RotateRadMatrix44('y', angle_y) 
        * vector3(-1.0f, 0, 0) * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_RIGHT))
    {
        lookat += RotateRadMatrix44('y', angle_y) 
        * vector3(1.0f, 0, 0) * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_DELETE))
    {
        angle_y -= 0.03f * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_PAGEDOWN))
    {
        angle_y += 0.03f * movement;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_HOME))
    {
        zoom /= 1.03f;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_END))
    {
        zoom *= 1.03f;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_INSERT))
    {
        angle_x -= 0.03f * movement;
        if (angle_x < 3.14f / 180.0f * -40.0f) angle_x = 3.14f / 180.0f * -40.0f;
    }
    if (Keyboard::Instance().KeyPressed(SDLK_PAGEUP))
    {
        angle_x += 0.03f * movement;
        if (angle_x > 3.14f / 180.0f * 80.0f) angle_x = 3.14f / 180.0f * 80.0f;
    }

    if (Keyboard::Instance().KeyPressed(SDLK_F1) && (Keyboard::Instance().KeyPressed(SDLK_LSHIFT) || Keyboard::Instance().KeyPressed(SDLK_RSHIFT)))
    {
        editor_world->Save(1);
    }
    else if (Keyboard::Instance().KeyPressed(SDLK_F1))
    {
        editor_world->Load(1);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_F2) && (Keyboard::Instance().KeyPressed(SDLK_LSHIFT) || Keyboard::Instance().KeyPressed(SDLK_RSHIFT)))
    {
        editor_world->Save(2);
    }
    else if (Keyboard::Instance().KeyPressed(SDLK_F2))
    {
        editor_world->Load(2);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_F3) && (Keyboard::Instance().KeyPressed(SDLK_LSHIFT) || Keyboard::Instance().KeyPressed(SDLK_RSHIFT)))
    {
        editor_world->Save(3);
    }
    else if (Keyboard::Instance().KeyPressed(SDLK_F3))
    {
        editor_world->Load(3);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_F4) && (Keyboard::Instance().KeyPressed(SDLK_LSHIFT) || Keyboard::Instance().KeyPressed(SDLK_RSHIFT)))
    {
        editor_world->Save(4);
    }
    else if (Keyboard::Instance().KeyPressed(SDLK_F4))
    {
        editor_world->Load(4);
    }

    camera = lookat + zoom * direction;

    direction = RotateRadMatrix44('y', angle_y) 
        * RotateRadMatrix44('x', angle_x)
        * vector3(0.0f, 1.0f, 1.0f).normalize();

    int mouse_x, mouse_y;
    bool ray_active = true;
    Mouse::Instance().getAbsoluteMouseState800600(&mouse_x, &mouse_y);

    static long minus_last_pressed;
    static long equals_last_pressed;

    if (Keyboard::Instance().KeyPressed(SDLK_MINUS))
    {
        if (GetTime() - minus_last_pressed > 100)
        {
            horizontal_plane -= 1.0f;

            minus_last_pressed = GetTime();
        }
    }
    if (Keyboard::Instance().KeyPressed(SDLK_EQUALS))
    {
        if (GetTime() - equals_last_pressed > 100)
        {
            horizontal_plane += 1.0f;

            equals_last_pressed = GetTime();
        }
    }



    vector3 origin, direction;
    mouse_pick_ray(perspective_matrix, view_matrix, mouse_x, mouse_y, 800, 600, origin, direction);
    editor_world->Ray(ray_active, origin, origin + direction.normalize() * 3000.0f, horizontal_plane);
}
/*
void Editor::DrawTransparentHorizontalPlane(float y)
{
    y--;

    IGS->EnableLighting(false);
    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, IdentityMatrix44());

    IGS->SetTexture(0, white_texture);

    vector3 surface[4];

    surface[0] = vector3(-1.0f,  0.0f,  1.0f);
    surface[1] = vector3(-1.0f,  0.0f, -1.0f);
    surface[2] = vector3( 1.0f,  0.0f, -1.0f);
    surface[3] = vector3( 1.0f,  0.0f,  1.0f);

    float scale = 256.0f;

    for (int n = 0; n < 4; n++)
    {
        surface[n].x *= scale;
        surface[n].y *= scale;
        surface[n].z *= scale;

        surface[n].x += 256.0f;
        surface[n].z += 256.0f;

        surface[n] -= vector3(2048.0f, -y, 2048.0f);
    }

    VERTEX_XYZ_COLOUR_UV1 points[4];

    points[0].u = 0.0f;
    points[0].v = 0.0f;
    points[0].colour = COLOUR(1.0f, 1.0f, 1.0f, 0.4f);

    points[1].u = 0.0f;
    points[1].v = 16.0f;
    points[1].colour = COLOUR(1.0f, 1.0f, 1.0f, 0.4f);

    points[2].u = 16.0f;
    points[2].v = 16.0f;
    points[2].colour = COLOUR(1.0f, 1.0f, 1.0f, 0.4f);

    points[3].u = 16.0f;
    points[3].v = 0.0f;
    points[3].colour = COLOUR(1.0f, 1.0f, 1.0f, 0.4f);

    unsigned short indices[] = { 0, 1, 2, 2, 3, 0};

    IGS->SetCullMode(CULL_NONE);
    IGS->SetBlendMode(SRC_ALPHA_ONE_MINUS_SRC_ALPHA);

    for (int i = 0; i < 10; i++)
    {
        for (int j = 0; j < 10; j++)
        {
            points[0].point = surface[0] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
            points[1].point = surface[1] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
            points[2].point = surface[2] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);
            points[3].point = surface[3] + (float)i * vector3(512.0f, 0.0f, 0.0f) + (float)j * vector3(0.0f, 0.0f, 512.0f);

            IGS->SetDepthBufferMode(READ_AND_WRITE);
            IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_COLOUR | VTX_UV1, (float*)points, 4, indices, 6);
        }
    }

    IGS->SetDepthBufferMode(READ_AND_WRITE);

    IGS->SetCullMode(CULL_CCW);
    IGS->SetBlendMode(BLEND_DISABLE);
}
*/











