// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "first_person.h"
#include "mouse.h"
#include "modelfactory.h"
#include "gettime.h"
#include "transform.h"
#include "object_renderer.h"
#include <iostream>
#include <math.h>

//! Constructor. Creates a geom and body for the sphere.
FirstPerson::FirstPerson(dWorldID world2, dSpaceID space2, vector3 initial_position,
    vector3 initial_direction, std::string launcher_directory,
    std::string launcher_filename, IGraphicsSystem *igs2)
{
    OdeThread::ode_mutex_lock();

    world = world2;
    space = space2;
    igs = igs2;

    pitch_angle = 0.00f;
    yaw_angle   = 0.01f;

    initial_direction.normalize();
    float x_dir = initial_direction.x;
    float y_dir = initial_direction.y;
    float z_dir = initial_direction.z;

    yaw_angle   = atan2(-x_dir, -z_dir) + 0.01f;

    position    = initial_position;
    direction   = initial_direction.normalize();

    // Create the sphere body
    dMass m;
    body_sphere = dBodyCreate(world);
    dBodySetPosition(body_sphere, position.x, position.y, position.z);
    dMassSetSphere (&m, 1, FPRADIUS);
    dMassAdjust (&m, FPMASS);
    dBodySetMass (body_sphere,&m);
    geom_sphere = dCreateSphere(space, FPRADIUS);
    dGeomSetBody(geom_sphere, body_sphere);

    OdeThread::ode_mutex_unlock();

/*
    // Load the 3d models for the grenade launcher.
    launcher = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
    launcher->Load(launcher_directory, launcher_filename, igs);
*/
    shotgun = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
    shotgun->Load(launcher_directory, launcher_filename, igs);
    shotgun_pose = shotgun->CreateAnimationInstance();

    smooth = 0.0;
}

FirstPerson::~FirstPerson()
{
    //delete launcher;
    delete shotgun;

    OdeThread::ode_mutex_lock();

    dBodyDestroy(body_sphere);
    dGeomDestroy(geom_sphere);
    
    OdeThread::ode_mutex_unlock();
}

//! Returns the camera position
void FirstPerson::GetCameraPosition(vector3 &camera, vector3 &lookat, vector3 &up)
{
    OdeThread::ode_mutex_lock();

    camera = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_sphere);
    camera += vector3(0.0, 4.0, 0.0); // make person taller.

    lookat = camera + (direction * 10.0f);

    up = vector3(0.0, 1.0, 0.0);

    position = camera;

    OdeThread::ode_mutex_unlock();
}

//! Returns the projectile spawn point and direction the player is aiming
void FirstPerson::GetWeaponLocationAndDirection(vector3& weapon_position, vector3& weapon_direction)
{
    OdeThread::ode_mutex_lock();

    weapon_direction = direction.normalize();
    weapon_position = transform(vector3(0.0f, 0.0f, 0.0f), this->GetTransformation()) + direction * 2.0f;

    OdeThread::ode_mutex_unlock();
}

//! Updates the player according to user input.
void FirstPerson::Update(int flags)
{
    OdeThread::ode_mutex_lock();

    int mouse_x, mouse_y;

    Mouse::Instance().getRelativeMouseState(&mouse_x, &mouse_y);

    yaw_angle   -= float(mouse_x) / 200.0f;
    pitch_angle -= float(mouse_y) / 200.0f;

    if (pitch_angle < -1.5f) pitch_angle = -1.5f;
    if (pitch_angle >  1.5f) pitch_angle =  1.5f;

    // Roate at_ point by angle_x radians
    float at_x =  0.0f;
    float at_z = -1.0f;
    float at_y =  0.0f;
    float at_z2 = (at_y*sin(pitch_angle))+(at_z*cos(pitch_angle));
    float at_y2 = (at_y*cos(pitch_angle))-(at_z*sin(pitch_angle));

    at_z =at_z2;
    at_y =at_y2;

    float at_x3 = (at_x*cos(yaw_angle)) + (at_z*sin(yaw_angle));
    float at_z3 =-(at_x*sin(yaw_angle)) + (at_z*cos(yaw_angle));

    at_x =at_x3;
    at_z =at_z3;

    direction.x = at_x;
    direction.y = at_y;
    direction.z = at_z;

/*
    const dReal *temp = dBodyGetAngularVel(body_sphere);

    vector3 vel(temp[0], temp[1], temp[2]);

    vector3 move(0.0, 0.0, 0.0);

    vector3 forwards(direction.z, 0, -direction.x);
    forwards.normalize();
    vector3 backwards = -forwards;

    vector3 left(forwards.z, 0, -forwards.x);
    vector3 right = -left;

    if (flags & FP_RUN_FORWARDS)
    {
        move += forwards;
    }
    if (flags & FP_RUN_BACKWARDS)
    {
        move += backwards;
    }
    if (flags & FP_STRAFE_LEFT)
    {
        move += left;
    }
    if (flags & FP_STRAFE_RIGHT)
    {
        move += right;
    }

    move.normalize();

    vector3 new_vel = vel / 1.5 + move / 1.5;


    dBodySetAngularVel(body_sphere, new_vel.x, new_vel.y, new_vel.z);
*/
    vector3 move(0.0, 0.0, 0.0);

    vector3 forwards(direction.z, 0, -direction.x);
    forwards.normalize();
    vector3 backwards = -forwards;

    vector3 left(forwards.z, 0, -forwards.x);
    vector3 right = -left;
    
    moving = false;

    if (flags & FP_RUN_FORWARDS)
    {
        move += forwards;
        moving = true;
    }
    if (flags & FP_RUN_BACKWARDS)
    {
        move += backwards;
        moving = true;
    }
    if (flags & FP_STRAFE_LEFT)
    {
        move += left;
        moving = true;
    }
    if (flags & FP_STRAFE_RIGHT)
    {
        move += right;
        moving = true;
    }

    move.normalize();

    vector3 new_vel = move * 3.0f;

    dBodySetAngularVel(body_sphere, new_vel.x, new_vel.y, new_vel.z);


    OdeThread::ode_mutex_unlock();
}

matrix44 FirstPerson::GetTransformation()
{
    // All this bob_x and bob_y stuff does is control how the weapon is held. It bobs up and down
    // a bit when running. When still it moves to a lesser extent. The blending between large and
    // small bobbing is done over 10 frames, so it's frame rate dependent but it doesn't matter
    // much because it didn't look too bad when there was no interpolated transition.

    if (moving == false)
    {
        if (smooth > 0.0) smooth -= 0.1f;
    }
    else
    {
        if (smooth < 1.0) smooth += 0.1f;
    }

    float bob_x_moving = 0.02 * sin((float(GetTime()%10000)/10000.0) * 3.14 * 2.0);
    float bob_y_moving = 0.02 * cos((float(GetTime()%12345)/12345.0) * 3.14 * 2.0);
    float bob_z_moving =-0.10f;

    float bob_x_static = 0.03 * sin((float(GetTime()%1000)/1000.0) * 3.14 * 2.0);
    float bob_y_static = 0.02 * cos((float(GetTime()%1234)/1234.0) * 3.14 * 2.0);
    float bob_z_static = 0.00f;

    float bob_x = bob_x_moving + (bob_x_static - bob_x_moving) * smooth;
    float bob_y = bob_y_moving + (bob_y_static - bob_y_moving) * smooth;
    float bob_z = bob_z_moving + (bob_z_static - bob_z_moving) * smooth;

    // Here is where the transformation matrix to draw the weapon is computed. At the moment it's
    // hardcoded for the grenade laucher but I'll fix this when adding the other weapons.
    transformation_weapon = IdentityMatrix44();

    transformation_weapon *= TranslateMatrix44(position.x, position.y, position.z);
    transformation_weapon *= TranslateMatrix44(direction.x * 2.0, direction.y * 2.0, direction.z * 2.0);
    transformation_weapon *= RotateRadMatrix44('Y', yaw_angle);
    transformation_weapon *= RotateRadMatrix44('X', pitch_angle);
    transformation_weapon *= TranslateMatrix44(0.3 + bob_x, -0.4 + bob_y, 0.9 + bob_z);
    transformation_weapon *= RotateRadMatrix44('Y', 3.14f / 2.0f);

    return transformation_weapon;
}

//! Renders the gun
void FirstPerson::Render(float frame_time)
{
    ObjectRenderer::Instance().AddModel(shotgun, this->GetTransformation(), frame_time, shotgun_pose);
}




