// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef FIRST_PERSON_H
#define FIRST_PERSON_H

#include "mtxlib.h"
#include "model.h"
#include "ode_thread.h"
#include <string>

// This class deals with the First Person camera and strafing and mouse looking and with jumping
// and the physics of the player's movement using ODE and dealing with the player as a sphere.

#define FPRADIUS    1.0 // radius of player sphere
#define FPMASS      0.2 // player mass

// Flags to describe the direction of the buggy
enum
{
    // If this flag is set person will run forwards
    FP_RUN_FORWARDS = 1,
    // If this flag is set person will run backwards
    FP_RUN_BACKWARDS= 2,
    // If this flag is set person will strafe left
    FP_STRAFE_LEFT  = 4,
    // If this flag is set person will strafe right
    FP_STRAFE_RIGHT = 8
};

class FirstPerson
{
public:

    //! Constructor. Creates a geom and body for the sphere.
    /*!
        Creates a sphere to represent the player's geometry.
    */
    FirstPerson(dWorldID world2, dSpaceID space2, vector3 initial_position,
        vector3 initial_direction, std::string launcher_directory, 
        std::string launcher_filename, IGraphicsSystem* igs2);
    ~FirstPerson();

    //! Returns the camera position
    /*!
        Returns the position and orientation of the first person camera.
        \param camera the position of the camera.
        \param lookat the point the camera is pointed towards.
        \param up the up vector of the camera.
    */
    void GetCameraPosition(vector3& camera, vector3& lookat, vector3& up);

    //! Returns the location of the player
    vector3 GetPosition() { return position; }
    
    //! Returns the projectile spawn point and direction the player is aiming
    void GetWeaponLocationAndDirection(vector3& weapon_position, vector3& weapon_direction);

    //! Updates the player according to user input.
    void Update(int flags);

    //! Renders the gun
    void Render(float frame_time = 0.0f);

    //! Gets the transformation of the gun to world space.
    matrix44 GetTransformation();

private:

    // The ODE variables necessary to create geoms and bodies.
    dWorldID world;
    dSpaceID space;

    // Body and geom for player's enclosing sphere.
    dBodyID body_sphere;
    dGeomID geom_sphere;

    vector3 position;   // position of the person.
    vector3 direction;  // lookat vector should always be of unit length.

    float pitch_angle;  // rotation around the X-axis, the axis running left and right.
    float yaw_angle;    // rotation around the Y-axis, the axis running up and down
    
    // Set to true if player is moving
    bool moving;
    // Slowly set to 0.0 if moving. 1.0 if not moving.
    float smooth;

    // A pointer to the rendering interface.
    IGraphicsSystem *igs;

    matrix44 transformation_weapon;

    IModel* shotgun;    // shotgun 3d model
    ANIMATION_INSTANCE_HANDLE shotgun_pose;
};

#endif // FIRST_PERSON_H

