// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include <iostream>
#include <algorithm>
#include <sstream>
#include "graphicssystemopengl.h"
#include "surface.h"

extern int g_width, g_height;    // width, height

static bool QueryExtension(char *extName)
{
    char* p = (char*) glGetString(GL_EXTENSIONS);
    char* end;
    if (p == NULL) return false;
    end = p + strlen(p);
    while (p < end)
    {
        unsigned int n = strcspn(p, " ");
        if ((strlen(extName)==n) && (strncmp(extName, p, n)==0))
        {
            return true;
        }
        p+= (n+1);
    }
    return false;
}

// Initialises the graphics system
GraphicsSystemOpenGL::GraphicsSystemOpenGL()
{
    for (int n = 0; n < 8; n++)
    {
        lights_enabled[n] = false;
    }

    // Set the viewport size
    glViewport(0, 0, g_width, g_height);

    // Check if some extensions are supported.
    stencil_wrap_available = QueryExtension("GL_EXT_stencil_wrap");
    arb_texture_env_dot3_available = QueryExtension("GL_ARB_texture_env_dot3");
    arb_texture_env_combine_available = QueryExtension("GL_ARB_texture_env_combine");
    ext_texture_filter_anisotropic_available = QueryExtension("GL_EXT_texture_filter_anisotropic");

    show_wireframe = false;
    show_shadow_volumes = false;

#if defined(WIN32)
    // -----------------------------------------------------------------------
    // Get pointers to the address of some GL functions
    // -----------------------------------------------------------------------
    glActiveTextureARB      = (PFNGLACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress ("glActiveTextureARB");
    glClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREPROC) SDL_GL_GetProcAddress ("glClientActiveTextureARB");
    glDrawRangeElements     = (PFNGLDRAWRANGEELEMENTSPROC) SDL_GL_GetProcAddress ("glDrawRangeElements");
    glTexImage3D            = (PFNGLTEXIMAGE3DPROC) SDL_GL_GetProcAddress ("glTexImage3D");
#endif

    vertex_buffers = new vertex_buffer[MAX_BUFFERS];
    index_buffers = new index_buffer[MAX_BUFFERS];
    textures = new texture[MAX_BUFFERS];
    cubemap_textures = new cubemap_texture[MAX_BUFFERS];
}

// Deinitialises the graphics system
GraphicsSystemOpenGL::~GraphicsSystemOpenGL()
{
    for (int v = 0; v < MAX_BUFFERS; v++)
    {
        if (vertex_buffers[v].buffer != NULL)
        {
            delete[] vertex_buffers[v].buffer;
            vertex_buffers[v].buffer = NULL;
        }
    }
    for (int i = 0; i < MAX_BUFFERS; i++)
    {
        if (index_buffers[i].buffer != NULL)
        {
            delete[] index_buffers[i].buffer;
            index_buffers[i].buffer = NULL;
        }
    }
    for (int t = 0; t < MAX_BUFFERS; t++)
    {
        textures[t].copy.Clear();
    }

    delete[] vertex_buffers;
    delete[] index_buffers;
    delete[] textures;
}

// Clears the screen
void GraphicsSystemOpenGL::Clear(COLOUR colour)
{
    // Clear the colour buffer and depth buffer
    glClearColor(colour.r, colour.g, colour.b, colour.a);
    glClearDepth(1.0);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void GraphicsSystemOpenGL::SetViewportSize(int width, int height)
{
    // Set the viewport size
    glViewport(0, 0, width, height);
}

// Begins the scene.
void GraphicsSystemOpenGL::Begin()
{
    // Set all the lights
    for (int n = 0; n < 8; n++)
    {
        if (lights_enabled[n] == true)
        {
            SetLight(n, lights[n]);
        }
    }
}

// Ends the scene.
void GraphicsSystemOpenGL::End()
{

}

// Displays the backbuffer onto the screen.
void GraphicsSystemOpenGL::Display()
{
    SDL_GL_SwapBuffers();
}

VERTEX_BUFFER_HANDLE GraphicsSystemOpenGL::CreateVertexList(int vertex_format, float* vertex_list,
    int number_of_vertices)
{
    int number_of_floats_per_vertex = GetNumberOfFloatsInFormat(vertex_format);

    VERTEX_BUFFER_HANDLE handle = 1;
    while(1)
    {
        if (vertex_buffers[handle].buffer == NULL)
        {
            int buffer_size = number_of_vertices * number_of_floats_per_vertex * sizeof(float);
            
            // Store this vertex buffer in the array of buffers.
            vertex_buffers[handle].buffer = new float[number_of_vertices * number_of_floats_per_vertex];
            memcpy(vertex_buffers[handle].buffer, vertex_list, buffer_size);

            // Store some information with it
            vertex_buffers[handle].number_of_vertices = number_of_vertices;
            vertex_buffers[handle].number_of_floats_per_vertex = number_of_floats_per_vertex;
            vertex_buffers[handle].vertex_format = vertex_format;

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Vertex Buffers reached");
        }
    }
}

//! Deletes a vertex buffer
void GraphicsSystemOpenGL::DeleteVertexList(VERTEX_BUFFER_HANDLE handle)
{
    if (vertex_buffers[handle].buffer != NULL)
    {
        delete[] vertex_buffers[handle].buffer;
        vertex_buffers[handle].buffer = NULL;
    }
}

//! Creates an index buffer. An array of indices.
INDEX_BUFFER_HANDLE GraphicsSystemOpenGL::CreateIndexBuffer(std::vector<unsigned short> index_list)
{
    // Loop through the vector handle array looking for a free slot
    INDEX_BUFFER_HANDLE handle = 1;

    while(1)
    {
        if (index_buffers[handle].buffer == NULL)
        {
            // Store the Index buffer
            index_buffers[handle].buffer = new unsigned short[index_list.size()];
            for (unsigned int i = 0; i < index_list.size(); i++)
            {
                index_buffers[handle].buffer[i] = index_list[i];
            }

            index_buffers[handle].number_of_indices = index_list.size();

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Index Buffers reached");
        }
    }
}

//! Deletes an index buffer.
void GraphicsSystemOpenGL::DeleteIndexBuffer(INDEX_BUFFER_HANDLE handle)
{
    if (index_buffers[handle].buffer != NULL)
    {
        delete[] index_buffers[handle].buffer;
        index_buffers[handle].buffer = NULL;
    }
}

//! Creates a texture map from an image file
TEXTURE_HANDLE GraphicsSystemOpenGL::CreateTexture(std::string texture_filename, bool generate_mipmaps)
{
    // Check if the texture is already loaded and if so use that instance and increment the ref. count.
    std::vector<texture_ref_count>::iterator t;
    for (t = texture_ref_counts.begin(); t != texture_ref_counts.end(); t++)
    {
        if (t->filename == texture_filename)
        {
            t->reference_count++;
            return t->handle;
        }
    }

    // Load the image from file.
    Surface texture(texture_filename);

    TEXTURE_HANDLE handle = CreateTexture(texture, generate_mipmaps);

    // Start a new reference count.
    texture_ref_count new_ref_count;
    new_ref_count.filename = texture_filename;
    new_ref_count.handle = handle;
    texture_ref_counts.push_back(new_ref_count);

    return handle;
}

//! Creates a texture map from a Surface.
TEXTURE_HANDLE GraphicsSystemOpenGL::CreateTexture(Surface surface, bool generate_mipmaps)
{
    GLuint texName;

    // Create the texture
    glGenTextures(1, &texName);

    CreateTexture(texName, surface, generate_mipmaps);

    // Loop through the textures array looking for a free slot
    TEXTURE_HANDLE handle = 1;

    while(1)
    {
        if (textures[handle].used == false)
        {
            // Store this texture in the array
            textures[handle].texture_id = texName;
            textures[handle].copy = surface;
            textures[handle].generate_mipmaps = generate_mipmaps;
            textures[handle].used = true;

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Textures reached");
        }
    }
}

// Creates a texture map from a Surface.
void GraphicsSystemOpenGL::CreateTexture(GLuint texName, Surface surface, bool generate_mipmaps)
{
    glBindTexture(GL_TEXTURE_2D, texName);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    if (generate_mipmaps)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    }
    else
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }

    // If anisotropy is supported, use it.
    if (ext_texture_filter_anisotropic_available)
    {
        GLfloat maxAnisotropy;
        glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAnisotropy);
        glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, &maxAnisotropy);
    }

    int max_width;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_width);

    if (surface.GetWidth() > max_width)
    {
        surface.MinFilter(max_width, max_width);
    }

    if (surface.GetSurfaceType() == R8G8B8)
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, surface.GetWidth(), surface.GetHeight(), 0, GL_RGB,
            GL_UNSIGNED_BYTE, surface.buffer);

        if (generate_mipmaps)
        {
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, surface.GetWidth(), surface.GetHeight(), GL_RGB,
                GL_UNSIGNED_BYTE, surface.buffer);
        }

    }
    else if (surface.GetSurfaceType() == R8G8B8A8)
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, surface.GetWidth(), surface.GetHeight(), 0, GL_RGBA,
            GL_UNSIGNED_BYTE, surface.buffer);

        if (generate_mipmaps)
        {
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, surface.GetWidth(), surface.GetHeight(), GL_RGBA,
                GL_UNSIGNED_BYTE, surface.buffer);
        }
    }
    else if (surface.GetSurfaceType() == X8)
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, surface.GetWidth(), surface.GetHeight(), 0, GL_ALPHA,
            GL_UNSIGNED_BYTE, surface.buffer);

        if (generate_mipmaps)
        {
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, surface.GetWidth(), surface.GetHeight(), GL_ALPHA,
                GL_UNSIGNED_BYTE, surface.buffer);
        }
    }
    else
    {
        throw std::string("Surface passed to CreateTexture has type other than R8G8B8 or R8G8B8A8");
    }
}

void GraphicsSystemOpenGL::CopyScreenToTexture(TEXTURE_HANDLE handle, int x, int y, int width, int height)
{
    glBindTexture(GL_TEXTURE_2D, textures[handle].texture_id);

    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, x, y, width, height, 0);
}

//! Deletes a texture
void GraphicsSystemOpenGL::DeleteTexture(TEXTURE_HANDLE handle)
{
    // Check if the texture has a ref. count
    std::vector<texture_ref_count>::iterator t;
    for (t = texture_ref_counts.begin(); t != texture_ref_counts.end(); t++)
    {
        if (t->handle == handle)
        {
            t->reference_count--;

            if (t->reference_count > 0)
            {
                return;
            }
            else
            {
                texture_ref_counts.erase(t);
                break;
            }
        }
    }

    GLuint texName = textures[handle].texture_id;
    glDeleteTextures(1, &texName);

    textures[handle].used = false;
    textures[handle].copy.Clear();
}


CUBEMAP_TEXTURE_HANDLE GraphicsSystemOpenGL::CreateCubemapTexture(Surface px, Surface nx, Surface py, Surface ny,
        Surface pz, Surface nz)
{
    GLuint texName;

    // Create the texture
    glGenTextures(1, &texName);

    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, texName);
    glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    int width = px.GetWidth();
    int height = px.GetHeight();

    glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, px.buffer);
    glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, py.buffer);
    glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, pz.buffer);
    glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, nx.buffer);
    glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, ny.buffer);
    glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0, 3, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, nz.buffer);

    // Loop through the textures array looking for a free slot
    CUBEMAP_TEXTURE_HANDLE handle = 1;

    while(1)
    {
        if (cubemap_textures[handle].used == false)
        {
            // Store this texture in the array
            cubemap_textures[handle].texture_id = texName;
            cubemap_textures[handle].copy_px = px;
            cubemap_textures[handle].copy_px = nx;
            cubemap_textures[handle].copy_px = py;
            cubemap_textures[handle].copy_px = ny;
            cubemap_textures[handle].copy_px = pz;
            cubemap_textures[handle].copy_px = nz;

            cubemap_textures[handle].used = true;

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of cubemap textures reached");
        }
    }
}

TEXTURE3D_HANDLE GraphicsSystemOpenGL::CreateTexture3D(unsigned char* buffer, int texture_size)
{
    GLuint texName;

    // Create the texture
    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_3D_EXT, texName);

    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage3D(GL_TEXTURE_3D_EXT, 0, GL_INTENSITY, texture_size, texture_size, texture_size,
        0, GL_LUMINANCE, GL_UNSIGNED_BYTE, buffer);

    // Loop through the textures array looking for a free slot
    TEXTURE3D_HANDLE handle = 1;

    while(1)
    {
        if (textures3D[handle].used == false)
        {
            // Store this texture in the array
            textures3D[handle].texture_id = texName;
            textures3D[handle].buffer_copy = NULL;  // TODO: Modify this for restoring 3D textures
            textures3D[handle].texture_size = texture_size;

            textures3D[handle].used = true;

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of 3D textures reached");
        }
    }
}

//! Restores all textures
void GraphicsSystemOpenGL::RestoreTextures()
{
    for (TEXTURE_HANDLE handle = 1; handle < MAX_BUFFERS; handle++)
    {
        if (textures[handle].used == true)
        {
            CreateTexture(textures[handle].texture_id, textures[handle].copy, textures[handle].generate_mipmaps);
        }
    }
}

//! Sets the texture for multitexturing stage.
void GraphicsSystemOpenGL::SetTexture(int stage, TEXTURE_HANDLE handle, texture_type type)
{
    if (stage == 0) glActiveTextureARB(GL_TEXTURE0_ARB);
    if (stage == 1) glActiveTextureARB(GL_TEXTURE1_ARB);
    if (stage == 2) glActiveTextureARB(GL_TEXTURE2_ARB);

    if (handle != 0)
    {
        if (type == TEXTURE_2D)
        {
            glDisable(GL_TEXTURE_CUBE_MAP_ARB);
            glDisable(GL_TEXTURE_3D_EXT);
            glEnable(GL_TEXTURE_2D);
        }
        else if (type == TEXTURE_CUBEMAP)
        {
            glDisable(GL_TEXTURE_2D);
            glDisable(GL_TEXTURE_3D_EXT);
            glEnable(GL_TEXTURE_CUBE_MAP_ARB);
        }
        else // if (type == TEXTURE_3D)
        {
            glDisable(GL_TEXTURE_2D);
            glDisable(GL_TEXTURE_CUBE_MAP_ARB);
            glEnable(GL_TEXTURE_3D_EXT);
        }
        glBindTexture(GL_TEXTURE_2D, textures[handle].texture_id);

    }
    else
    {
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_TEXTURE_CUBE_MAP_ARB);
        glDisable(GL_TEXTURE_3D_EXT);
    }
}

//! Sets the mode for automatic texture generation
void GraphicsSystemOpenGL::SetUVGenerationMode(int stage, UV_GENERATION_MODE mode)
{
    if (stage == 0) glActiveTextureARB(GL_TEXTURE0_ARB);
    if (stage == 1) glActiveTextureARB(GL_TEXTURE1_ARB);
    if (stage == 2) glActiveTextureARB(GL_TEXTURE2_ARB);

    if (mode == UV_SPHERE_MAPPING)
    {
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    }
    else if (mode == UV_OBJECT_SPACE)
    {
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glEnable(GL_TEXTURE_GEN_R);
        glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        glTexGenf(GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        GLfloat sgenparams[] = {1.0f, 0.0f, 0.0f, 0.0f};
        glTexGenfv(GL_S, GL_OBJECT_PLANE, sgenparams);
        GLfloat tgenparams[] = {0.0f, 1.0f, 0.0f, 0.0f};
        glTexGenfv(GL_T, GL_OBJECT_PLANE, tgenparams);
        GLfloat rgenparams[] = {0.0f, 0.0f, 1.0f, 0.0f};
        glTexGenfv(GL_R, GL_OBJECT_PLANE, rgenparams);
    }
    else // mode == UV_DEFAULT)
    {
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glDisable(GL_TEXTURE_GEN_R);
    }
}

//! Sets the texture blend mode.
void GraphicsSystemOpenGL::SetMultiTextureMode(TEXTURE_BLEND_MODE mode)
{
    if (mode == NORMAL)
    {
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }
    if (mode == DETAIL_TEXTURE)
    {
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

        glActiveTextureARB(GL_TEXTURE1_ARB);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
        glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2);
    }
    if (mode == DOT3_BUMP)
    {
        throw std::string("DOT3 not implemented");
/*
        // UNIT 0
        // find dot product of N (stored in the texture map) and L (stored
        // as the PRIMARY_COLOR).
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_DOT3_RGB_ARB);

        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);

        // UNIT 1
        // modulate the base texture by N.L
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);

        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
*/
    }
}

//! Sets the texture wrapping mode
void GraphicsSystemOpenGL::SetTextureWrappingMode(int stage, TEXTURE_MODE mode)
{
    if (stage == 0) glActiveTextureARB(GL_TEXTURE0_ARB);
    if (stage == 1) glActiveTextureARB(GL_TEXTURE1_ARB);
    if (stage == 2) glActiveTextureARB(GL_TEXTURE2_ARB);

    if (mode == CLAMP_TEXTURE)
    {
        // In OpenGL these functions only apply to the currently bound texture
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    else if (mode == WRAP_TEXTURE)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    else if (mode == CLAMP_TEXTURE_S)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    }
    else if (mode == CLAMP_TEXTURE_T)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    else if (mode == CLAMP_TEXTURE_R)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
    }
    else if (mode == WRAP_TEXTURE_S)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    }
    else if (mode == WRAP_TEXTURE_T)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    else // if (mode == WRAP_TEXTURE_R)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_REPEAT);
    }
}


//! Sets the culling mode
void GraphicsSystemOpenGL::SetCullMode(cull_mode cull)
{
    switch(cull)
    {
        case CULL_CW:
            glFrontFace(GL_CCW);
            glCullFace(GL_BACK);
            glEnable(GL_CULL_FACE);
            break;
        case CULL_CCW:
            glFrontFace(GL_CW);
            glCullFace(GL_BACK);
            glEnable(GL_CULL_FACE);
            break;
        case CULL_NONE:
            glDisable(GL_CULL_FACE);
            break;
        default:
            break;
    };
}

int GraphicsSystemOpenGL::GetNumberOfFloatsInFormat(const int vertex_format)
{
    int floats_in_format = 0;

    if (vertex_format & VTX_XYZ) floats_in_format += 3;
    if (vertex_format & VTX_NORMAL) floats_in_format += 3;
    if (vertex_format & VTX_COLOUR) floats_in_format += 4;
    if (vertex_format & VTX_UV1) floats_in_format += 2;
    if (vertex_format & VTX_UV2) floats_in_format += 2;
    if (vertex_format & VTX_UV3) floats_in_format += 2;

    return floats_in_format;
}

void GraphicsSystemOpenGL::DisableAllArrays()
{
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE2_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

void GraphicsSystemOpenGL::EnableNeededArrays(const int vertex_format, const int number_of_floats_per_vertex,
    float* vertex_array)
{
    int floats_offset = 0;
    int floats_in_format = number_of_floats_per_vertex;

    if (vertex_format & VTX_XYZ)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 3;
    }
    if (vertex_format & VTX_NORMAL)
    {
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 3;
    }
    if (vertex_format & VTX_COLOUR)
    {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 4;
    }
    if (vertex_format & VTX_UV1)
    {
        glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
    if (vertex_format & VTX_UV2)
    {
        glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
    if (vertex_format & VTX_UV3)
    {
        glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
}


void GraphicsSystemOpenGL::DrawTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count)
{
    int floats_in_format = vertex_buffers[vhandle].number_of_floats_per_vertex;
    int vertex_format = vertex_buffers[vhandle].vertex_format;
    float* vertex_array;

    vertex_array = vertex_buffers[vhandle].buffer;

    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawArrays(GL_TRIANGLES, 0, vertex_count);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}


void GraphicsSystemOpenGL::DrawIndexedTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle,
    int vertex_count, INDEX_BUFFER_HANDLE ihandle, int index_count)
{
    int floats_in_format = vertex_buffers[vhandle].number_of_floats_per_vertex;
    int vertex_format = vertex_buffers[vhandle].vertex_format;
    float* vertex_array;

    vertex_array = vertex_buffers[vhandle].buffer;

    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawRangeElements(GL_TRIANGLES, 0, vertex_count, index_count,
            GL_UNSIGNED_SHORT, index_buffers[ihandle].buffer);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

void GraphicsSystemOpenGL::DrawTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
    int vertex_count)
{
    int floats_in_format = GetNumberOfFloatsInFormat(vertex_format);
    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawArrays(GL_TRIANGLES, 0, vertex_count);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

void GraphicsSystemOpenGL::DrawIndexedTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
    int vertex_count, unsigned short *index_array, int index_count)
{
    int floats_in_format = GetNumberOfFloatsInFormat(vertex_format);
    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawRangeElements(GL_TRIANGLES, 0, vertex_count, index_count,
            GL_UNSIGNED_SHORT, index_array);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

matrix44 obj_to_world, view, projection;

//! Sets a transformation matrix.
void GraphicsSystemOpenGL::SetTransformationMatrix(
    TRANSFORMATION_TYPE type, const matrix44& m)
{
    GLfloat matrix[16];
    matrix[0] = m.col[0].x;
    matrix[1] = m.col[0].y;
    matrix[2] = m.col[0].z;
    matrix[3] = m.col[0].w;

    matrix[4] = m.col[1].x;
    matrix[5] = m.col[1].y;
    matrix[6] = m.col[1].z;
    matrix[7] = m.col[1].w;

    matrix[8] = m.col[2].x;
    matrix[9] = m.col[2].y;
    matrix[10]= m.col[2].z;
    matrix[11]= m.col[2].w;

    matrix[12] = m.col[3].x;
    matrix[13] = m.col[3].y;
    matrix[14] = m.col[3].z;
    matrix[15] = m.col[3].w;

    if (type == OBJECT_TO_WORLD_TRANSFORM)
    {
        obj_to_world = m;

        glMatrixMode(GL_MODELVIEW);
        glLoadMatrixf(view_matrix);
        glMultMatrixf(matrix);
    }
    else if (type == VIEW_TRANSFORM)
    {
        view = m;

        // Copy matrix to view_matrix
        memcpy(view_matrix, matrix, sizeof(GLfloat) * 16);
    }
    else if (type == PROJECTION_TRANSFORM)
    {
        projection = m;

        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf(matrix);
    }
    else throw std::string("Unsupported type passed to SetTransformationMatrix");
}

//! Returns a transformation matrix.
matrix44 GraphicsSystemOpenGL::GetTransformationMatrix(TRANSFORMATION_TYPE type)
{
    if (type == OBJECT_TO_WORLD_TRANSFORM) return obj_to_world;
    else if (type == VIEW_TRANSFORM) return view;
    else if (type == PROJECTION_TRANSFORM) return projection;
    else throw std::string("Unsupported type passed to GetTransformationMatrix");
}

//! Sets which material is used for lighting.
void GraphicsSystemOpenGL::SetMaterial(MATERIAL material)
{
    GLfloat mat_diffuse[] = {material.diffuse.r, material.diffuse.g,
        material.diffuse.b, material.diffuse.a };
    
    GLfloat mat_specular[] = { material.specular.r, material.specular.g,
        material.specular.b, material.specular.a };
        
    GLfloat mat_emissive[] = { material.emissive.r, material.emissive.g,
        material.emissive.b, material.emissive.a };

    GLfloat mat_shininess[] = { material.shiny * 128.0 };

    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_EMISSION, mat_emissive);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
}

//! Sets a light
void GraphicsSystemOpenGL::SetLight(int light_number, LIGHT light)
{
    // Copy the light to the array of LIGHT structures.
    lights[light_number] = light;

    int LIGHT;

    if (light_number == 0)      LIGHT = GL_LIGHT0;
    else if (light_number == 1) LIGHT = GL_LIGHT1;
    else if (light_number == 2) LIGHT = GL_LIGHT2;
    else if (light_number == 3) LIGHT = GL_LIGHT3;
    else if (light_number == 4) LIGHT = GL_LIGHT4;
    else if (light_number == 5) LIGHT = GL_LIGHT5;
    else if (light_number == 6) LIGHT = GL_LIGHT6;
    else if (light_number == 7) LIGHT = GL_LIGHT7;
    else throw std::string("8 lights maximum has been exceeded");

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadMatrixf(view_matrix);

    GLfloat light_direction[] = {   -lights[light_number].direction.x,
                                    -lights[light_number].direction.y,
                                    -lights[light_number].direction.z,
                                    0.0f };

    GLfloat light_ambient[] = {     0.0,
                                    0.0,
                                    0.0,
                                    1.0};
    GLfloat light_diffuse[] = {     lights[light_number].diffuse.r,
                                    lights[light_number].diffuse.g,
                                    lights[light_number].diffuse.b,
                                    1.0};
    GLfloat light_specular[]= {     lights[light_number].specular.r,
                                    lights[light_number].specular.g,
                                    lights[light_number].specular.b,
                                    1.0};

    float cutoff_angle = 30.0f;

    glLightfv(LIGHT, GL_POSITION,   light_direction);

    glLightfv(LIGHT, GL_AMBIENT,    light_ambient);
    glLightfv(LIGHT, GL_DIFFUSE,    light_diffuse);
    glLightfv(LIGHT, GL_SPECULAR,   light_specular);

    glEnable(GL_LIGHTING);
    glEnable(LIGHT);

    glPopMatrix();
}

//! Enables/Disables lighting.
void GraphicsSystemOpenGL::EnableLighting(bool enable)
{
    if (enable)
    {
        glEnable(GL_LIGHTING);
    }
    else
    {
        glDisable(GL_LIGHTING);
    }
}

//! Enables or disables a light
void GraphicsSystemOpenGL::EnableLight(int light_number, bool enable)
{
    int LIGHT;

    if (light_number == 0)      LIGHT = GL_LIGHT0;
    else if (light_number == 1) LIGHT = GL_LIGHT1;
    else if (light_number == 2) LIGHT = GL_LIGHT2;
    else if (light_number == 3) LIGHT = GL_LIGHT3;
    else if (light_number == 4) LIGHT = GL_LIGHT4;
    else if (light_number == 5) LIGHT = GL_LIGHT5;
    else if (light_number == 6) LIGHT = GL_LIGHT6;
    else if (light_number == 7) LIGHT = GL_LIGHT7;
    else throw std::string("8 lights maximum has been exceeded");
    
    if (enable == true)
    {
        glEnable(LIGHT);
        lights_enabled[light_number] = true;
    }
    else if (enable == false)
    {
        glDisable(LIGHT);
        lights_enabled[light_number] = true;
    }
}

void GraphicsSystemOpenGL::SetColour(COLOUR colour)
{
    glColor4f(colour.r, colour.g, colour.b, colour.a);
}

//! Sets the fog parameters
void GraphicsSystemOpenGL::FogParams(float density, COLOUR colour)
{
/*
    // Set the Fog type to GL_EXP.
    glFogi(GL_FOG_MODE, GL_EXP);

    // Set the fog colour
    GLfloat fog_colour[] = {colour.r, colour.g, colour.b };

    glFogfv(GL_FOG_COLOR, fog_colour);
    glFogf(GL_FOG_DENSITY, density);
    glHint(GL_FOG_HINT, GL_DONT_CARE);
*/
/*
    // Set the Fog type to GL_LINEAR
    glFogi(GL_FOG_MODE, GL_LINEAR);

    // Set the fog colour
    GLfloat fog_colour[] = {colour.r, colour.g, colour.b };

    glFogfv(GL_FOG_COLOR, fog_colour);
    glFogf(GL_FOG_START, 850.0f);
    glFogf(GL_FOG_END, 920.0f);
*/
}

//! Enables or disables fog
void GraphicsSystemOpenGL::EnableFog(bool enable)
{
    if (enable == true)
    {
        glEnable(GL_FOG);
    }
    else
    {
        glDisable(GL_FOG);
    }
}

//! Sets the blend mode
void GraphicsSystemOpenGL::SetBlendMode(blend_mode blend)
{
    if (blend == SRC_ALPHA_ONE)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    }
    else if (blend == SRC_ALPHA_ONE_MINUS_SRC_ALPHA)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    else if (blend == ONE_ONE)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE);
    }
    else if (blend == ONE_ONE_MINUS_SRC_COLOR)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
    }
    else
    {
        glDisable(GL_BLEND);
    }
}

//! Sets the depth buffer mode
void GraphicsSystemOpenGL::SetDepthBufferMode(depth_buffer_mode depth)
{
    if (depth == READ_ONLY)
    {
        glEnable(GL_DEPTH_TEST);
        glDepthMask(GL_FALSE);
        glDepthFunc(GL_LEQUAL);
    }
    else if (depth == READ_AND_WRITE)
    {
        glEnable(GL_DEPTH_TEST);
        glDepthMask(GL_TRUE);
        glDepthFunc(GL_LEQUAL);
    }
    else if (depth == IGNORE_DEPTH)
    {
        glDepthFunc(GL_ALWAYS);
        glDepthMask(GL_FALSE);
    }
}

//! Enables or disables wireframe mode
void GraphicsSystemOpenGL::ShowWireframe(bool enable)
{
    if (enable && show_wireframe == false)
    {
        glPolygonMode(GL_FRONT, GL_LINE);
        glPolygonMode(GL_BACK, GL_LINE);

        show_wireframe = true;
    }
    else if (!enable && show_wireframe == true)
    {
        glPolygonMode(GL_FRONT, GL_FILL);
        glPolygonMode(GL_BACK, GL_FILL);

        show_wireframe = false;
    }
}

//! Enables or disables rendering shadow volumes to colour buffer mode
void GraphicsSystemOpenGL::ShowShadowVolumes(bool enable)
{
    show_shadow_volumes = enable;
}

//! Grabs a screenshot and saves it to file.
void GraphicsSystemOpenGL::Screenshot(std::string filename, bool plus_alpha)
{
    unsigned char* pixels = new unsigned char[g_width * g_height * 4];

    glPixelStorei(GL_PACK_ALIGNMENT, 4);
    glPixelStorei(GL_PACK_ROW_LENGTH, 0);
    glPixelStorei(GL_PACK_SKIP_ROWS, 0);
    glPixelStorei(GL_PACK_SKIP_PIXELS, 0);

    if (plus_alpha == true)
    {
        glReadPixels(0, 0, g_width, g_height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

        Surface screenshot(g_width, g_height, R8G8B8A8);
        for (int idx = 1; idx <= g_height; idx++)
        {
            memcpy(screenshot.buffer + 4 * g_width * (idx - 1), pixels + 4 * g_width * (g_height - idx), 4 * g_width);
        }

        screenshot.SavePNG(filename);
    }
    else // don't read alpha
    {
        glReadPixels(0, 0, g_width, g_height, GL_RGB, GL_UNSIGNED_BYTE, pixels);

        Surface screenshot(g_width, g_height, R8G8B8);
        for (int idx = 1; idx <= g_height; idx++)
        {
            memcpy(screenshot.buffer + 3 * g_width * (idx - 1), pixels + 3 * g_width * (g_height - idx), 3 * g_width);
        }

        screenshot.Save(filename);
    }

    delete[] pixels;
}

void GraphicsSystemOpenGL::ClearStencil()
{
    glClear(GL_STENCIL_BUFFER_BIT);
}

void GraphicsSystemOpenGL::PrepareStencil()
{
    glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_STENCIL_BUFFER_BIT);
    glDisable(GL_LIGHTING);                                         // Turn off lighting
    glDepthMask(GL_FALSE);                                          // Turn off writing to the depth buffer
    glEnable(GL_STENCIL_TEST);                                      // Turn on stencil buffer testing

    if (show_shadow_volumes)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE);

        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);            // Draw into the colour buffer
    }
    else
    {
        glDisable(GL_BLEND);
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);        // Don't draw into the colour buffer
    }

    glStencilFunc(GL_ALWAYS, 0, ~0);
    glStencilMask(~0);
    glDepthFunc(GL_LESS);

    SetTexture(0, 0);
    SetTexture(1, 0);
}

void GraphicsSystemOpenGL::StencilSetFirstShadowPass(shadow_mode mode)
{
    if (mode == Z_PASS_SHADOWS)
    {
        // First pass. Increase stencil value in the shadow
        SetCullMode(CULL_CW);
        glStencilOp(GL_KEEP, GL_KEEP, stencil_wrap_available ? GL_INCR_WRAP_EXT : GL_INCR);

        if (show_shadow_volumes)
        {
            glColor4f(0.0f, 0.2f, 0.0f, 0.0f);  // green
        }
    }
    else // (mode == Z_FAIL_SHADOWS)
    {
        // First pass.
        SetCullMode(CULL_CCW);
        glStencilOp(GL_KEEP, stencil_wrap_available ? GL_INCR_WRAP_EXT : GL_INCR, GL_KEEP);

        if (show_shadow_volumes)
        {
            glColor4f(0.2f, 0.0f, 0.0f, 0.0f);  // red
        }
    }
}

void GraphicsSystemOpenGL::StencilSetSecondShadowPass(shadow_mode mode)
{
    if (mode == Z_PASS_SHADOWS)
    {
        // Second pass. Decrease stencil value in the shadow
        SetCullMode(CULL_CCW);
        glStencilOp(GL_KEEP, GL_KEEP, stencil_wrap_available ? GL_DECR_WRAP_EXT : GL_DECR);
    }
    else // (mode == Z_FAIL_SHADOWS)
    {
        // Second pass.
        SetCullMode(CULL_CW);
        glStencilOp(GL_KEEP, stencil_wrap_available ? GL_DECR_WRAP_EXT : GL_DECR, GL_KEEP);
    }
}

void GraphicsSystemOpenGL::StencilTestBegin()
{
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);    // Enable rendering to colour buffer for all components
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);

    glStencilFunc(GL_EQUAL, 0, ~0);
    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
}

void GraphicsSystemOpenGL::StencilTestEnd()
{
    glPopAttrib();
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);    // Enable rendering to colour buffer for all components
    glDepthFunc(GL_LEQUAL);
}

// Draws a sprite. The texture to use must be set already before calling this function.
void GraphicsSystemOpenGL::Draw2D(VERTEX_XYZ_UV1 top_left, VERTEX_XYZ_UV1 top_right,
    VERTEX_XYZ_UV1 bottom_left, VERTEX_XYZ_UV1 bottom_right, COLOUR c)
{
    SetBlendMode(SRC_ALPHA_ONE_MINUS_SRC_ALPHA);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    glOrtho(0, 800, 600, 0, -1, 1);
    glDisable(GL_DEPTH_TEST);
    glColor4f(c.r, c.g, c.b, c.a);

    glBegin(GL_QUADS);
    glTexCoord2f(top_left.u, top_left.v);           glVertex2f(top_left.point.x, top_left.point.y);
    glTexCoord2f(top_right.u, top_right.v);         glVertex2f(top_right.point.x, top_right.point.y);
    glTexCoord2f(bottom_left.u, bottom_left.v);     glVertex2f(bottom_left.point.x, bottom_left.point.y);
    glTexCoord2f(bottom_right.u, bottom_right.v);   glVertex2f(bottom_right.point.x, bottom_right.point.y);
    glEnd();

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glEnable(GL_DEPTH_TEST);
    glPopMatrix();
}



