// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef GRAPHICS_SYSTEM_OPENGL_H
#define GRAPHICS_SYSTEM_OPENGL_H

/*!
    \file graphicssystemopengl.h
    \brief OpenGL interface class.
    File containing interface for OpenGL which inherits from the abstract
    interface.
*/

#include "graphicssystem.h"

#if defined(WIN32)
#include <windows.h>
#endif // defined(WIN32)

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#define MAX_BUFFERS 1000

//! OpenGL interface class
/*!
    Interface for OpenGL, inherits from the abstract interface.
*/
class GraphicsSystemOpenGL : public IGraphicsSystem
{
public:
    //! Initialises the graphics system
    GraphicsSystemOpenGL();

    //! Deinitialises the graphics system
    virtual ~GraphicsSystemOpenGL();

    //! Clears the screen
    virtual void Clear(COLOUR colour);


    virtual void SetViewportSize(int width, int height);

    //! Begins the scene.
    virtual void Begin();

    //! Ends the scene.
    virtual void End();

    //! Displays the backbuffer onto the screen.
    virtual void Display();

    virtual VERTEX_BUFFER_HANDLE CreateVertexList(int vertex_format,
        float* vertex_list, int number_of_vertices);

    //! Deletes a vertex buffer
    virtual void DeleteVertexList(VERTEX_BUFFER_HANDLE handle);

    //! Creates an index buffer. An array of indices.
    virtual INDEX_BUFFER_HANDLE CreateIndexBuffer(std::vector<unsigned short> index_list);

    //! Deletes an index buffer.
    virtual void DeleteIndexBuffer(INDEX_BUFFER_HANDLE handle);

    //! Creates a texture map
    virtual TEXTURE_HANDLE CreateTexture(std::string texture_filename, bool generate_mipmaps);

    //! Creates a texture map from a Surface.
    virtual TEXTURE_HANDLE CreateTexture(Surface surface, bool generate_mipmaps);

    virtual void CopyScreenToTexture(TEXTURE_HANDLE handle, int x, int y, int width, int height);

    //! Deletes a texture
    virtual void DeleteTexture(TEXTURE_HANDLE handle);

    virtual CUBEMAP_TEXTURE_HANDLE CreateCubemapTexture(Surface px, Surface nx, Surface py, Surface ny,
        Surface pz, Surface nz);

    virtual TEXTURE3D_HANDLE CreateTexture3D(unsigned char* buffer, int texture_size);

    //! Restores all textures
    virtual void RestoreTextures();

    //! Sets the texture for multitexturing stage.
    virtual void SetTexture(int stage, TEXTURE_HANDLE handle, texture_type type = TEXTURE_2D);

    //! Sets the mode for automatic texture generation
    virtual void SetUVGenerationMode(int stage, UV_GENERATION_MODE mode);

    //! Sets the texture blend mode.
    virtual void SetMultiTextureMode(TEXTURE_BLEND_MODE mode);

    //! Sets the texture wrapping mode
    virtual void SetTextureWrappingMode(int stage, TEXTURE_MODE mode);

    //! Sets the culling mode
    virtual void SetCullMode(cull_mode cull);

    //! Sets which material is used for lighting.
    virtual void SetMaterial(MATERIAL material);

    //! Draws triangles from a vertex buffer.
    virtual void DrawTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count);

    //! Draws triangles from a vertex buffer and index buffer.
    virtual void DrawIndexedTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count,
        INDEX_BUFFER_HANDLE ihandle, int index_count);

    //! Draws triangles from a vertex array
    virtual void DrawTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count);

    //! Draws triangles from a vertex array and index array
    virtual void DrawIndexedTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count, unsigned short *index_array, int index_count);

    //! Sets a light
    virtual void SetLight(int light_number, LIGHT light);

    //! Enables/Disables lighting.
    virtual void EnableLighting(bool enable);

    //! Enables or disables a light
    virtual void EnableLight(int light_number, bool enable);

    virtual void SetColour(COLOUR colour);

    //! Sets a transformation matrix.
    virtual void SetTransformationMatrix(TRANSFORMATION_TYPE type, const matrix44&);

    //! Returns a transformation matrix.
    virtual matrix44 GetTransformationMatrix(TRANSFORMATION_TYPE type);

    //! Sets the fog parameters
    virtual void FogParams(float density, COLOUR colour);

    //! Enables or disables fog
    virtual void EnableFog(bool enable);

    //! Sets the blend mode
    virtual void SetBlendMode(blend_mode blend);

    //! Sets the depth buffer mode
    virtual void SetDepthBufferMode(depth_buffer_mode depth);

    //! Enables or disables wireframe mode
    virtual void ShowWireframe(bool enable);

    //! Enables or disables rendering shadow volumes to colour buffer mode
    virtual void ShowShadowVolumes(bool enable);

    //! Grabs a screenshot and saves it to file.
    virtual void Screenshot(std::string filename, bool plus_alpha = false);

    virtual void ClearStencil();

    virtual void PrepareStencil();

    virtual void StencilSetFirstShadowPass(shadow_mode mode);

    virtual void StencilSetSecondShadowPass(shadow_mode mode);

    virtual void StencilTestBegin();

    virtual void StencilTestEnd();

    // Draws a sprite. The texture to use must be set already before calling this function.
    virtual void Draw2D(VERTEX_XYZ_UV1 top_left, VERTEX_XYZ_UV1 top_right, VERTEX_XYZ_UV1 bottom_left,
        VERTEX_XYZ_UV1 bottom_right, COLOUR c);

private:

    void CreateTexture(GLuint texName, Surface surface, bool generate_mipmaps);

    int GetNumberOfFloatsInFormat(const int vertex_format);
    void DisableAllArrays();
    void EnableNeededArrays(const int vertex_format, const int number_of_floats_per_vertex, float* vertex_array);

    // Matrix containing the viewing transformation
    GLfloat view_matrix[16];

    // A group of lights.
    LIGHT lights[8];
    bool lights_enabled[8];

#if defined(WIN32)
    PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
    PFNGLCLIENTACTIVETEXTUREPROC glClientActiveTextureARB;
    PFNGLDRAWRANGEELEMENTSPROC glDrawRangeElements;
    PFNGLTEXIMAGE3DPROC glTexImage3D;
#endif // defined(WIN32)

    class vertex_buffer
    {
    public:
        vertex_buffer::vertex_buffer() { buffer = NULL; }

        float* buffer;
        int number_of_vertices;
        int number_of_floats_per_vertex;
        int vertex_format;
    };

    vertex_buffer* vertex_buffers;

    class index_buffer
    {
    public:
        index_buffer::index_buffer() { buffer = NULL; }

        unsigned short* buffer;
        int number_of_indices;
    };

    index_buffer* index_buffers;

    class texture
    {
    public:
        texture::texture() { this->used = false; }
        bool used;
        GLuint texture_id;
        Surface copy;
        bool generate_mipmaps;
    };

    texture* textures;

    class texture_ref_count
    {
    public:
        texture_ref_count::texture_ref_count() { this->reference_count = 1; }
        int reference_count;
        std::string filename;
        TEXTURE_HANDLE handle;
    };

    std::vector<texture_ref_count> texture_ref_counts;

    class cubemap_texture
    {
    public:
        cubemap_texture::cubemap_texture() { this->used = false; }
        bool used;
        GLuint texture_id;
        Surface copy_px;
        Surface copy_nx;
        Surface copy_py;
        Surface copy_ny;
        Surface copy_pz;
        Surface copy_nz;
    };

    cubemap_texture* cubemap_textures;

    class texture3D
    {
    public:
        texture3D::texture3D() { this->used = false; }
        bool used;
        GLuint texture_id;
        unsigned char* buffer_copy;
        int texture_size;
    };

    texture3D* textures3D;

    // booleans set if some extensions are available;
    bool stencil_wrap_available;
    bool arb_texture_env_combine_available;
    bool arb_texture_env_dot3_available;
    bool ext_texture_filter_anisotropic_available;


    bool show_shadow_volumes;
    bool show_wireframe;
};

#endif
