// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef KEYBOARD_H
#define KEYBOARD_H

/*!
    \file keyboard.h
    \brief Handles the keyboard.
*/

#include "SDL.h"
#include "SDL_keysym.h"
#include <vector>

//! The number of keys to use for the array of key_states
#define NUMBER_OF_KEYS 512

//! Class that handles input from the keyboard.
class Keyboard
{
public:
    //! Returns a reference to the one and only instance of this class.
    static Keyboard& Instance()
    {
        static Keyboard Obj;
        return Obj;
    }

    //! Checks if a key is currently pressed.
    /*!
        \param key SDLKey code of the key to check e.g. SDLK_a
        \return Returns true if the key corresponding to key is down. Returns false otherwise.
    */
    bool KeyPressed(SDLKey key);

    //! Queries the current value of a switch corresponding to a key
    /*!
        A switch is a boolean value that is set to true when a key is pressed and then set to false
        if the key is pressed again.
        \param key SDLKey code of the key to check e.g. SDLK_a
        \return The current value of the boolean switch.
    */
    bool QueryCurrentValueOfSwitch(SDLKey key);

    //! To be called whenever an SDL_KEYUP or SDL_KEYDOWN event is received by the main loop.
    /*!
        This function should only be called by the main event loop in sdlopengl.cpp
        \param key Pointer to an SDL_KeyboardEvent.
    */
    void KeyEventNotify(SDL_KeyboardEvent* key);

private:

    // Constructor is private in this singleton
    Keyboard();

    // Array containing status of all the keys on the keyboard.
    bool key_state[NUMBER_OF_KEYS];

    // Array containing a list of switches that can be toggled
    bool switch_state[NUMBER_OF_KEYS];
};

#endif // KEYBOARD_H
