// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "level.h"
#include "filenames.h"
#include "error.h"
#include "keyboard.h"
#include "mouse.h"
#include "get_distance.h"
#include <iostream>
#include "camera.h"
#include "object_renderer.h"
#include "gettime.h"
#include "application.h"
#include "hud_renderer.h"
#include "hud.h"
#include "vector_math.h"
#include "lighting.h"
#include <sstream>

extern TEXTURE_HANDLE font_texture01;

Camera cam1(int( 500 * 1.5));
Camera cam2(int(1000 * 1.5));

// Constructor
Level::Level(IGraphicsSystem* IGS, Application* application)
{
    // Store the pointer to the rendering interface.
    this->IGS = IGS;
    this->app = application;

    try
    {
        world = OdeThread::Instance().getWorld();
        space = OdeThread::Instance().getSpace();

        fp = new FirstPerson(world, space, vector3(25, 40, 25), vector3(0, 0, -1),
            data_dir() + slash(), "shotgun.txt", IGS);

        buggy = new Buggy(world, space, IGS, data_dir() + slash() + "buggy" + slash(),
            "buggy_chassis.txt", "wheel4.txt", vector3(20, 40, 20));

        // Load a smoke texture with an alpha channel.
        smoke0_texture= IGS->CreateTexture(data_dir() + slash() + "smoke0.png", true);

        level_world = new World(IGS);
        level_world->Load(data_dir() + slash() + "world" + slash() + "world.txt");

        //GenerateLightmap3D(IGS);

        OdeThread::Instance().start_thread();
    }
    catch (std::string error_message)
    {
        Error(error_message);
    }
}

// Destructor
Level::~Level()
{
    //IGS->Screenshot("screenie.png", true);

    OdeThread::Instance().stop_thread();

    delete buggy;

    delete level_world;

    if (fp != NULL) delete fp;

    IGS->DeleteTexture(smoke0_texture);
}

// Main update function. Call this once per frame.
void Level::Update()
{
#if !defined(PHYSICS_IN_SEPARATE_THREAD)
    OdeThread::Instance().Update();
#endif // !defined(PHYSICS_IN_SEPARATE_THREAD)

    static bool control_vehicle = false;
    static bool return_key_up = true;
    if (!Keyboard::Instance().KeyPressed(SDLK_RETURN)) return_key_up = true;

    if (Keyboard::Instance().KeyPressed(SDLK_RETURN) && control_vehicle && return_key_up)
    {
        dWorldID world = OdeThread::Instance().getWorld();
        dSpaceID space = OdeThread::Instance().getSpace();

        vector3 spawn_point = buggy->GetPointInWorldSpace(vector3(-8.0, 1.0, -5.0));

        fp = new FirstPerson(world, space, spawn_point,
            buggy->GetDirection(), data_dir() + slash(), "shotgun.txt", IGS);

        control_vehicle = false;
        return_key_up = false;
    }
    else if (Keyboard::Instance().KeyPressed(SDLK_RETURN) && return_key_up)
    {
        if (get_distance(buggy->GetPosition(), fp->GetPosition()) < 10.0f)
        {
            // Some camera trickery to smoothly interpolate between being on foot and
            // then overlooking the car
            vector3 camera, lookat, up;
            fp->GetCameraPosition(camera, lookat, up);
            cam1.clear();
            cam2.clear();
            cam2.setCurrentPosition(camera);
            cam1.setCurrentPosition(lookat);

            delete fp;
            fp = NULL;

            control_vehicle = true;
            return_key_up = false;
        }
    }

    dWorldID world = OdeThread::Instance().getWorld();
    dSpaceID space = OdeThread::Instance().getSpace();

    if (Keyboard::Instance().KeyPressed(SDLK_r))
    {
        delete buggy;
        buggy = new Buggy(world, space, IGS, data_dir() + slash() + "buggy" +
            slash(), "buggy_chassis.txt", "wheel4.txt", vector3(20, 40, 20));
    }

    if (control_vehicle == true)
    {
        int buggy_direction_flags = 0;

        if (Keyboard::Instance().KeyPressed(SDLK_UP) || Keyboard::Instance().KeyPressed(SDLK_w))
        {
            buggy_direction_flags |= BUGGY_FORWARD;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_DOWN) || Keyboard::Instance().KeyPressed(SDLK_s))
        {
            buggy_direction_flags |= BUGGY_REVERSE;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_LEFT) || Keyboard::Instance().KeyPressed(SDLK_a))
        {
            buggy_direction_flags |= BUGGY_TURN_LEFT;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_RIGHT) || Keyboard::Instance().KeyPressed(SDLK_d))
        {
            buggy_direction_flags |= BUGGY_TURN_RIGHT;
        }

        buggy->Update(buggy_direction_flags);
    }
    else
    {
        int buggy_direction_flags = 0;
        buggy->Update(buggy_direction_flags);

        int fp_direction_flags = 0;

        if (Keyboard::Instance().KeyPressed(SDLK_UP) || Keyboard::Instance().KeyPressed(SDLK_w))
        {
            fp_direction_flags |= FP_RUN_FORWARDS;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_DOWN) || Keyboard::Instance().KeyPressed(SDLK_s))
        {
            fp_direction_flags |= FP_RUN_BACKWARDS;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_LEFT) || Keyboard::Instance().KeyPressed(SDLK_a))
        {
            fp_direction_flags |= FP_STRAFE_LEFT;
        }
        if (Keyboard::Instance().KeyPressed(SDLK_RIGHT) || Keyboard::Instance().KeyPressed(SDLK_d))
        {
            fp_direction_flags |= FP_STRAFE_RIGHT;
        }

        fp->Update(fp_direction_flags);
    }

    if (control_vehicle == true)
    {
        buggy->GetCameraPosition(camera, lookat, up, true);

        cam2.setCurrentPosition(camera);
        camera = cam2.getSmooothedPosition();

        cam1.setCurrentPosition(lookat);
        lookat = cam1.getSmooothedPosition();
    }
    else
    {
        fp->GetCameraPosition(camera, lookat, up);
    }
}

//! Main render function. Call this once per frame.
void Level::Render1(vector3 camera)
{
    try
    {
        // Render the buggy
        buggy->Render();

        // Render the gun
        if (fp != NULL)
        {
            fp->Render(/*(GetTime()%1000)/50.0f*/0.0f);
        }

        level_world->Render();

/*
        for (float i = 0.0f; i < 5.0f; i += 1.0f)
        {
            for (float j = 0.0f; j < 5.0f; j += 1.0f)
            {
                for (float k = 0.0f; k < 5.0f; k += 1.0f)
                {
                    vector3 p = vector3(i*2.0f +164.0f, j*2.0f+30.0f, k*2.0f+358.0f);
                    float angle = (float)(GetTime() % (360 * 25)) / 25.0f;
                    ObjectRenderer::Instance().AddParticle(2.0f, COLOUR(0.8f, 0.8f, 0.8f, 0.4f), SRC_ALPHA_ONE_MINUS_SRC_ALPHA, smoke0_texture, p, angle);
                }
            }
        }
*/
    }
    catch (std::string error_message)
    {
        Error(error_message);
    }
}

std::string ConvertToTime(long milliseconds)
{
    std::stringstream time;

    time << ((milliseconds/100000)%10) << ((milliseconds/10000)%10) << ((milliseconds/1000)%10)
        << '.' << ((milliseconds/100)%10) << ((milliseconds/10)%10);// << (milliseconds%10);


    return time.str();
}

// Main render function. Call this once per frame.
void Level::Render2(vector3 camera)
{
    try
    {
        static long start_time = 0;
        static long started = false;
        static long best = 0;
        static long elapsed;

/*
        IGS->EnableFog(false);
        sky->Render(camera);
        sea->Render();
        IGS->EnableFog(true);
*/
    }
    catch (std::string error_message)
    {
        Error(error_message);
    }
}

