// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "lighting.h"
#include <math.h>
#include <sstream>

#define CUBEMAP_TEX_SIZE    256
#define SIZE_LIGHTMAP       512
#define SIZE_LIGHTMAP_3D    128

CUBEMAP_TEXTURE_HANDLE GenerateCubemap(IGraphicsSystem* IGS)
{
#define PackFloatInByte(in) (unsigned char)((((in) + 1.0f) / 2.0f) * 255.0f)
    int x, y, z;

    float map_size = (float)CUBEMAP_TEX_SIZE;
    float half_map_size = (float)(CUBEMAP_TEX_SIZE / 2);

    Surface cubemap_px(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);
    Surface cubemap_nx(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);
    Surface cubemap_py(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);
    Surface cubemap_ny(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);
    Surface cubemap_pz(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);
    Surface cubemap_nz(CUBEMAP_TEX_SIZE, CUBEMAP_TEX_SIZE, R8G8B8);


    // Positive X face
    for (y = 0; y < CUBEMAP_TEX_SIZE; y++)
    {
        for (z = 0; z < CUBEMAP_TEX_SIZE; z++)
        {
            float fX = float(half_map_size);
            float fY = float(half_map_size - y);
            float fZ = float(half_map_size - z);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_px.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 0] = PackFloatInByte(fX);
            cubemap_px.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 1] = PackFloatInByte(fY);
            cubemap_px.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

    // Negative X face
    for (y = 0; y < CUBEMAP_TEX_SIZE; y++)
    {
        for (z = 0; z < CUBEMAP_TEX_SIZE; z++)
        {
            float fX = float(-half_map_size);
            float fY = float(half_map_size - y);
            float fZ = float(z - half_map_size);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_nx.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 0] = PackFloatInByte(fX);
            cubemap_nx.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 1] = PackFloatInByte(fY);
            cubemap_nx.buffer[(y * CUBEMAP_TEX_SIZE + z) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

    // Positive Y face
    for (z = 0; z < CUBEMAP_TEX_SIZE; z++)
    {
        for (x = 0; x < CUBEMAP_TEX_SIZE; x++)
        {
            float fX = float(x - half_map_size);
            float fY = float(half_map_size);
            float fZ = float(z - half_map_size);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_py.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 0] = PackFloatInByte(fX);
            cubemap_py.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 1] = PackFloatInByte(fY);
            cubemap_py.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

    // Negative Y face
    for (z = 0; z < CUBEMAP_TEX_SIZE; z++)
    {
        for (x = 0; x < CUBEMAP_TEX_SIZE; x++)
        {
            float fX = float(x - half_map_size);
            float fY = float(-half_map_size);
            float fZ = float(half_map_size - z);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_ny.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 0] = PackFloatInByte(fX);
            cubemap_ny.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 1] = PackFloatInByte(fY);
            cubemap_ny.buffer[(z * CUBEMAP_TEX_SIZE + x) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

    // Positive Z face
    for (y = 0; y < CUBEMAP_TEX_SIZE; y++)
    {
        for (x = 0; x < CUBEMAP_TEX_SIZE; x++)
        {
            float fX = float(x - half_map_size);
            float fY = float(half_map_size - y);
            float fZ = float(half_map_size);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_pz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 0] = PackFloatInByte(fX);
            cubemap_pz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 1] = PackFloatInByte(fY);
            cubemap_pz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

    // Negative Z face
    for (y = 0; y < CUBEMAP_TEX_SIZE; y++)
    {
        for (x = 0; x < CUBEMAP_TEX_SIZE; x++)
        {
            float fX = float(half_map_size - x);
            float fY = float(half_map_size - y);
            float fZ = float(-half_map_size);

            float oolen = 1.0f / sqrt(fX*fX + fY*fY + fZ*fZ);
            fX *= oolen; fY *= oolen; fZ *= oolen;

            cubemap_nz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 0] = PackFloatInByte(fX);
            cubemap_nz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 1] = PackFloatInByte(fY);
            cubemap_nz.buffer[(y * CUBEMAP_TEX_SIZE + x) * 3 + 2] = PackFloatInByte(fZ);
        }
    }

/*
    cubemap_px.Save("px.png");
    cubemap_nx.Save("nx.png");
    cubemap_py.Save("py.png");
    cubemap_ny.Save("ny.png");
    cubemap_pz.Save("pz.png");
    cubemap_nz.Save("nz.png");
*/
    return IGS->CreateCubemapTexture(cubemap_px, cubemap_nx, cubemap_py, cubemap_ny, cubemap_pz, cubemap_nz);
}
/*
TEXTURE_HANDLE GenerateSpotlightmap(IGraphicsSystem* IGS)
{
    float rad = 255.0f;
    float rad_sq = rad*rad;

    Surface lightmap(SIZE_LIGHTMAP, SIZE_LIGHTMAP, X8);

    for (int t = 0; t < SIZE_LIGHTMAP; t++)
    {
        for (int s = 0; s < SIZE_LIGHTMAP; s++)
        {
            float d_sq = (s-SIZE_LIGHTMAP/2)*(s-SIZE_LIGHTMAP/2) + (t-SIZE_LIGHTMAP/2)*(t-SIZE_LIGHTMAP/2);
            if (d_sq < rad_sq)
            {
                float falloff = ((rad_sq - d_sq)/rad_sq) * ((rad_sq - d_sq)/rad_sq);
                lightmap.buffer[t*SIZE_LIGHTMAP + s] = (unsigned char)(255.0 * falloff);
            }
            else
            {
                lightmap.buffer[t*SIZE_LIGHTMAP + s] = 0;
            }
        }
    }

    //lightmap.Save("lightmap.png");

    return IGS->CreateTexture(lightmap, true);
}
*/

TEXTURE3D_HANDLE GenerateLightmap3D(IGraphicsSystem* IGS)
{
    unsigned char lightmap[SIZE_LIGHTMAP_3D * SIZE_LIGHTMAP_3D * SIZE_LIGHTMAP_3D];

    float rad = SIZE_LIGHTMAP_3D/2 - 1.0;
    float rad_sq = rad*rad;

    #define SQ(A) ((A)*(A))
    for (int r = 0; r < SIZE_LIGHTMAP_3D; r++)
    {
        for (int t = 0; t < SIZE_LIGHTMAP_3D; t++)
        {
            for (int s = 0; s < SIZE_LIGHTMAP_3D; s++)
            {
                float d_sq = SQ(r-SIZE_LIGHTMAP_3D/2) + SQ(s-SIZE_LIGHTMAP_3D/2) + SQ(t-SIZE_LIGHTMAP_3D/2);

                if (d_sq < rad_sq)
                {
                    float falloff = SQ((rad_sq - d_sq)/rad_sq);
                    lightmap[r*SIZE_LIGHTMAP_3D*SIZE_LIGHTMAP_3D + t*SIZE_LIGHTMAP_3D + s] = (unsigned char)(255.0 * falloff);
                }
                else
                {
                    lightmap[r*SIZE_LIGHTMAP_3D*SIZE_LIGHTMAP_3D + t*SIZE_LIGHTMAP_3D + s] = 0;
                }
            }
        }
    }
/*
    for (int slice = 0; slice < SIZE_LIGHTMAP_3D; slice++)
    {
        Surface lightmap_slice(SIZE_LIGHTMAP_3D, SIZE_LIGHTMAP_3D, X8);
        memcpy(lightmap_slice.buffer, lightmap + SIZE_LIGHTMAP_3D * SIZE_LIGHTMAP_3D * slice, SIZE_LIGHTMAP_3D * SIZE_LIGHTMAP_3D);
        std::stringstream filename;
        filename << "slice" << slice << ".png";
        lightmap_slice.Save(filename.str());
    }
*/
    return IGS->CreateTexture3D(lightmap, SIZE_LIGHTMAP_3D);
/*
  // Setup texture environment and load cubemap.
  glGenTextures(NUM_LIGHTMAP_TEXTURES, lightmapID);
  glBindTexture(GL_TEXTURE_3D_EXT, lightmapID[0]);
  glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S,
                  GL_MIRROR_CLAMP_TO_EDGE_ATI);
  glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T,
                  GL_MIRROR_CLAMP_TO_EDGE_ATI);
  glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT,
                  GL_MIRROR_CLAMP_TO_EDGE_ATI);
  glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage3DEXT(GL_TEXTURE_3D_EXT, 0, GL_INTENSITY,
                  SIZE_LIGHTMAP_3D, SIZE_LIGHTMAP_3D, SIZE_LIGHTMAP_3D,
                  0, GL_LUMINANCE, GL_UNSIGNED_BYTE, lightmap);
*/
}










