// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include "SDL_keysym.h"
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include "application.h"
#include "keyboard.h"
#include "filenames.h"
#include "settings.h"

int g_width;
int g_height;
int g_bpp       = 32;
int g_flags     = SDL_OPENGL | SDL_FULLSCREEN;

// Information about the current video settings
const SDL_VideoInfo* g_info = NULL;

int main(int argc, char* argv[])
{
    // First initialise SDL
    if (SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO) < 0)
    {
        std::cout << "SDL initialisation failed: " << SDL_GetError() << std::endl;
        exit(1);
    }

    // Shutdown SDL at exit
    atexit(SDL_Quit);

    // Let's get some video information
    g_info = SDL_GetVideoInfo();
    if (!g_info)
    {
        std::cout << "Video Query failed: " << SDL_GetError() << std::endl;
        exit(1);
    }
    int current_bpp = g_info->vfmt->BitsPerPixel;

    // Now we want to setup our requested window attributes for out OpenGL
    // window. We want *at least* 5 bits of red, green and blue. We also
    // want at least a 15-bit depth buffer.
    //
    // The last thing we do is request a double buffered window. '1' turns
    // on double buffering '0' turns it off.
    //
    // Note that we do not use SDL_DOUBLEBUF in the flags to
    // SDL_SetVideoMode. That does not affect the GL attribute state only
    // the standard 2D blitting setup.

    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 15);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);

    // Get the current settings from file.
    Settings::Instance().Load();
    g_width = Settings::Instance().screen_width;
    g_height= Settings::Instance().screen_height;

    // Set the video mode
    if (SDL_SetVideoMode(g_width, g_height, g_bpp, g_flags) == 0)
    {
        g_bpp = current_bpp;

        // Failed. Try again with current bpp.
        if (SDL_SetVideoMode(g_width, g_height, current_bpp, g_flags) == 0)
        {
            std::cout << "Failed to set video mode: " << SDL_GetError() << std::endl;
            exit(1);
        }
    }

    g_width = SDL_GetVideoSurface()->w;
    g_height= SDL_GetVideoSurface()->h;

    // Set the window caption
    SDL_WM_SetCaption("slickworm", "slickworm");

    // Hide the mouse cursor
    SDL_ShowCursor(0);

    // Grab the mouse and keyboard.
    SDL_WM_GrabInput(SDL_GRAB_ON);

/*
    // Enable Unicode translation
    SDL_EnableUNICODE( 1 );
*/

    Application App;


    bool exit = false;
    while(!exit)
    {
        // Call the application's update function
        App.Update();

        // Poll for keypresses
        SDL_Event event;
        while (SDL_PollEvent(&event))
        {
            // We are only worried about SDL_KEYDOWN events
            switch(event.type)
            {
                case SDL_KEYDOWN:
                case SDL_KEYUP:
                        if (event.key.keysym.sym == SDLK_ESCAPE) exit = 1;
                        Keyboard::Instance().KeyEventNotify(&event.key);
                    break;

                case SDL_QUIT:
                    exit = true;
                    break;

                default:
                    break;
            }
        }
    }

    // Because SDL_Quit was passed as a parameter to atexit() I don't need to call it here.
    return 0;
}
