// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "object_renderer.h"
#include "get_distance.h"
#include "transform.h"
#include <iostream>
#include <algorithm>

ObjectRenderer::ObjectRenderer()
{
    this->IGS = NULL;
}

ObjectRenderer::~ObjectRenderer()
{

}

void ObjectRenderer::Begin(IGraphicsSystem* IGS, vector3& camera, matrix44 view_matrix)
{
    this->IGS = IGS;
    this->camera = camera;
    this->view_matrix = view_matrix;

    transparent_group.clear();
    objects.clear();
    lights.clear();

    Light ambient;
    ambient.type = Light::AMBIENT;
    ambient.colour = COLOUR(0.0f, 0.0f, 0.0f, 1.0f);
    lights.push_back(ambient);
}

// Adds a model to the scene
void ObjectRenderer::AddModel(IModel* model, matrix44 transformation_model, float time, ANIMATION_INSTANCE_HANDLE instance)
{
    Object obj;
    obj.object            = model;
    obj.anim_time         = time;
    obj.anim_updated      = false;
    obj.anim_instance     = instance;
    obj.transformation    = transformation_model;

    objects.push_back(obj);
}

//! Adds a light to the scene
/*!
    Call this to add a light to the scene.
*/
void ObjectRenderer::AddLight(Light light)
{
    if (light.type == Light::AMBIENT)
    {
        lights[0].colour = light.colour;
    }
    else if (light.type == Light::DIRECTIONAL)
    {
        light.direction.normalize();

        if (light.direction.y < 0.0)
        {
            light.clip1 = vector3(0.0,      0.0,    0.0);
            light.clip2 = vector3(100.0,    0.0,    100.0);
            light.clip3 = vector3(100.0,    0.0,    0.0);
        }
        else if (light.direction.y > 0.0)
        {
            light.clip1 = vector3(0.0,      200.0,  0.0);
            light.clip2 = vector3(100.0,    200.0,  100.0);
            light.clip3 = vector3(100.0,    200.0,  0.0);
        }

        lights.push_back(light);
    }
    else
    {
        lights.push_back(light);
    }
}

// Adds a particle to the group of particles to be sorted and rendered.
void ObjectRenderer::AddParticle(float size, COLOUR particle_colour, blend_mode mode, TEXTURE_HANDLE texture,
    vector3 position, float angle)
{
    // from (http://nate.scuzzy.net/docs/billboard/)
    // Get the transpose of the view matrix
    matrix44 transposed = TransposeMatrix44(view_matrix);
    vector3 x = vector3(transposed.col[0].x, transposed.col[0].y, transposed.col[0].z);
    vector3 y = vector3(transposed.col[1].x, transposed.col[1].y, transposed.col[1].z);

    // Rotate x & y
    vector3 z = CrossProduct(x, y);
    matrix44 r = RotateRadMatrix44(z, DegToRad(angle));
    x = transform(x, r);
    y = transform(y, r);

    VERTEX_XYZ_COLOUR_UV1 vertices[4];
    vertices[0].u = 0.0f;
    vertices[0].v = 1.0f;
    vertices[0].colour.r = particle_colour.r;
    vertices[0].colour.g = particle_colour.g;
    vertices[0].colour.b = particle_colour.b;

    vertices[1].u = 0.0f;
    vertices[1].v = 0.0f;
    vertices[1].colour.r = particle_colour.r;
    vertices[1].colour.g = particle_colour.g;
    vertices[1].colour.b = particle_colour.b;

    vertices[2].u = 1.0f;
    vertices[2].v = 0.0f;

    vertices[2].colour.r = particle_colour.r;
    vertices[2].colour.g = particle_colour.g;
    vertices[2].colour.b = particle_colour.b;

    vertices[3].u = 1.0f;
    vertices[3].v = 1.0f;
    vertices[3].colour.r = particle_colour.r;
    vertices[3].colour.g = particle_colour.g;
    vertices[3].colour.b = particle_colour.b;

    vector3 a = position + ((-x - y) * size);
    vector3 b = position + ((-x + y) * size);
    vector3 c = position + (( x + y) * size);
    vector3 d = position + (( x - y) * size);

    for (int i = 0; i < 4; i++)
    {
        vertices[i].colour.a = particle_colour.a;
    }

    vertices[0].point = a;
    vertices[1].point = b;
    vertices[2].point = c;
    vertices[3].point = d;

    TransparentPolygon p1, p2;

    float distance = get_distance(camera, position);

    memcpy(p1.vertices, vertices, 3 * sizeof(VERTEX_XYZ_COLOUR_UV1));

    p1.distance = distance;
    p1.texture = texture;
    p1.mode = mode;

    p2.vertices[0] = vertices[2];
    p2.vertices[1] = vertices[3];
    p2.vertices[2] = vertices[0];
    p2.distance = distance;
    p2.texture = texture;
    p2.mode = mode;

    transparent_group.push_back(p1);
    transparent_group.push_back(p2);
}

// Renders all the objects in the scene
void ObjectRenderer::RenderObjects(Light current_light)
{
    // Loop through all the objects
    std::vector<Object>::iterator itor;
    for (itor = objects.begin(); itor != objects.end(); itor++)
    {
        IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, itor->transformation);

        // Determine if the object is within the frustum
        if (itor->object->IsWithinFrustum(itor->transformation))
        {
            if (itor->anim_instance != NULL)
            {
                IModel* model = (IModel*)itor->object;

                model->Update(itor->anim_time, itor->anim_instance);
                itor->anim_updated = true;
            }

            itor->object->Render(itor->anim_instance, current_light, itor->transformation);
        }
    }
}

// Render all the shadows of the objects in the scene
void ObjectRenderer::RenderShadows(Light current_light)
{
    IGS->PrepareStencil();

    // Loop through all the objects
    std::vector<Object>::iterator itor;
    for (itor = objects.begin(); itor != objects.end(); itor++)
    {
        // Determine whether or not the shadow is visible
        if (itor->object->IsShadowWithinFrustum(itor->transformation, current_light))
        {
            if (itor->anim_instance != NULL && itor->anim_updated == false)
            {
                IModel* model = (IModel*)itor->object;

                model->Update(itor->anim_time, itor->anim_instance);
                itor->anim_updated = true;
            }

            // Determine whether or not the camera is in the shadow volume
            shadow_mode mode = Z_PASS_SHADOWS;

            if (itor->object->IsCameraWithinShadow(itor->transformation, camera, current_light))
            {
                mode = Z_FAIL_SHADOWS;
            }

            itor->object->cast_shadow(IGS, itor->transformation, current_light, itor->anim_instance, mode);
        }
    }
}

// Renders all the objects and shadows in the scene.
void ObjectRenderer::RenderScene()
{

    IGS->SetBlendMode(BLEND_DISABLE);

    IGS->EnableLighting(false);
    IGS->SetColour(lights[0].colour);
    RenderObjects(lights[0]);

    // Render the scene for each light
    for (unsigned int current_light = 1; current_light < lights.size(); current_light++)
    {
        IGS->EnableLighting(false);

        // Clear the stencil buffer
        if (current_light > 1)
        {
            IGS->ClearStencil();
        }

        // Draw shadows
        RenderShadows(lights[current_light]);

        IGS->StencilTestBegin();

        IGS->SetBlendMode(ONE_ONE);

        // Set the light
        LIGHT light;
        memset(&light, 0, sizeof(LIGHT));
        light.type = DIRECTIONAL_LIGHT;
        light.direction = lights[current_light].direction;
        light.diffuse = lights[current_light].colour;
        IGS->SetLight(0, light);
        IGS->EnableLight(0, true);
        IGS->EnableLighting(true);

        RenderObjects(lights[current_light]);

        IGS->StencilTestEnd();
    }
}


// Renders all the particles
void ObjectRenderer::RenderTransparentObjects()
{
    IGS->SetCullMode(CULL_NONE);

    // Check if the number of particles is 0 and return if so.
    if (transparent_group.size() == 0) return;

    // Sort the transparent polygons by distance
    std::sort(transparent_group.begin(), transparent_group.end());


    IGS->SetDepthBufferMode(READ_ONLY);

    // Set the object to world transform to the identity
    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, IdentityMatrix44());

    // Loop through all the transparent polygons and render them
    std::vector<TransparentPolygon>::iterator itor;
    for (itor = transparent_group.begin(); itor != transparent_group.end(); itor++)
    {
        // Make sure the correct texture is Set.
        IGS->SetTexture(0, itor->texture);
        IGS->SetTextureWrappingMode(0, CLAMP_TEXTURE);
        // Make sure the correct blend mode is set.
        IGS->SetBlendMode(itor->mode);

        IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ | VTX_COLOUR | VTX_UV1, (float*)itor->vertices, 3);
    }

    IGS->SetBlendMode(BLEND_DISABLE);
    IGS->SetDepthBufferMode(READ_AND_WRITE);
}

