// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef OBJECT_RENDERER_H
#define OBJECT_RENDERER_H

#include "transform.h"
#include "graphicssystem.h"
#include "model.h"

/*!
    \file object_renderer.h
    \brief File containing Object renderer class.
*/

//! A structure holding info about a light in the scene.
class Light
{
public:
    enum TYPE
    {
        AMBIENT,
        DIRECTIONAL,
        POINT,
        SPOT
    };

    TYPE type;
    vector3 position;
    vector3 direction;
    COLOUR colour;

    // Clip plane if type == DIRECTIONAL
    vector3 clip1, clip2, clip3;
};

//! Object Renderer
class ObjectRenderer
{
public:
    //! Returns a reference to the one and only instance of this class.
    static ObjectRenderer& Instance()
    {
        static ObjectRenderer Obj;
        return Obj;
    }

    //! Destructor
    ~ObjectRenderer();

    //! Call this once per frame before calling any of the other functions.
    /*!
        Sets up some variables and clears the group of transparent objects.
        \param IGS Pointer to renderer interface
        \param camera Camera position in the scene
        \param view_matrix The matrix for transforming the scene into camera space.
    */
    void Begin(IGraphicsSystem *IGS, vector3& camera, matrix44 view_matrix);

    //! Adds a model to the scene
    /*!
        Call this to add a model to the scene. Models should be added to the scene every frame.
        \param model The model to add to the scene
        \param transformation_model The transformation of the model from local to world space.
        \param time Used if the model is skeletally animated
        \param instance Current animation instance if the model is skeletally animated
    */
    void AddModel(IModel* model, matrix44 transformation_model, float time = 0.0f, ANIMATION_INSTANCE_HANDLE instance = NULL);

    //! Adds a light to the scene
    /*!
        Call this to add a light to the scene.
    */
    void AddLight(Light light);

    // Adds a particle to the group of particles to be sorted and rendered.
    /*!
        Adds a particle to the scene. Particle gets added to a list of transparent objects that get sorted
        by distance from camera and then rendered.
        \param size Size of the particle
        \param particle_colour Alpha value used in blending plus rgb value used to colour the particle
        \param mode blend_mode used to render partcle.
        \param texture Particle texture
        \param position Position of the particle in the scene.
        \param angle the angle (in degrees) the particle is rotated in screen space.
    */
    void AddParticle(float size, COLOUR particle_colour, blend_mode mode, TEXTURE_HANDLE texture, vector3 position,
        float angle = 0.0f);

    // Renders all the objects and shadows in the scene.
    void RenderScene();

    //! Renders all the particles
    /*!
        Sorts all the transparent objects by distance from camera and then renders them.
    */
    void RenderTransparentObjects();

private:

    // Constructor is private in this singleton
    ObjectRenderer();

    //! Renders all the objects in the scene
    /*!
        Renders all the non transparent objects
    */
    void RenderObjects(Light current_light);

    //! Render all the shadows of the objects in the scene
    /*!
        Renders each of the shadows in turn.
    */
    void RenderShadows(Light current_light);

    // A pointer to the rendering interface
    IGraphicsSystem* IGS;

    // The camera position
    vector3 camera;

    //  The view matrix
    matrix44 view_matrix;

    //! A structure holding all the info needed to render a transparent triangle polygon
    class TransparentPolygon
    {
    public:
        VERTEX_XYZ_COLOUR_UV1 vertices[3];  // The 3 vertices of the trinagle.
        float distance;                     // The distance of the triangle from the camera
        TEXTURE_HANDLE texture;             // Texture used to render the particle.
        blend_mode mode;                    // Determines the blend mode

        // comparison operator. this will be called by std::sort and will sort them by distance
        bool operator<(const TransparentPolygon& other) const
        {
            return (this->distance > other.distance);
        }
    };

    // Our group of particles to be sorted.
    std::vector<TransparentPolygon> transparent_group;

    //! A structure holding info needed to render a shadowable object.
    class Object
    {
    public:
        Shadowable* object;
        float anim_time;
        bool anim_updated;
        ANIMATION_INSTANCE_HANDLE anim_instance;
        matrix44 transformation;
    };

    // The group of objects
    std::vector<Object> objects;

    // The group of lights
    std::vector<Light> lights;
};

#endif // OBJECT_RENDERER_H

