// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "model.h"
#include "transform.h"
#include "gettime.h"
#include "object_renderer.h"
#include <iostream>
#include <vector>

Shadowable::Shadowable()
{
    Vertices = NULL;
    Indices = NULL;
}

Shadowable::~Shadowable()
{
    if (Vertices != NULL)
    {
        delete[] Vertices;
        delete[] Indices;
        dGeomTriMeshDataDestroy(TriMeshData);
    }
}

void Shadowable::build_connectivity_list()
{
    vertices = GetVertices();
    indices = GetIndices();
/*
    Timer timer;
    timer.Go();

    std::cout << vertices.size() << " ";
*/
    faces.resize(indices.size() / 3);
    for (unsigned int f = 0; f < faces.size(); f++)
    {
        faces[f].vertex_indices[0] = indices[f * 3 + 0];
        faces[f].vertex_indices[1] = indices[f * 3 + 1];
        faces[f].vertex_indices[2] = indices[f * 3 + 2];

        for (int edge = 0; edge < 3; edge++)
        {
            faces[f].neighbour_indices[edge] = -1;
        }
    }

    //  for each face (A) in the object
    //      for each edge in A
    //          if we don't know this edges neighbour yet
    //              for each face (B) in the object (except A)
    //                  for each edge in B
    //                      if A's edge is the same as B's edge, then they are neighbouring each other on that edge
    //                          set the neighbour property for each face A and B, then move onto next edge in A

    unsigned int face_a, face_b, edge_a, edge_b;
    for (face_a = 0; face_a < faces.size(); face_a++)
    {
        for (edge_a = 0; edge_a < 3; edge_a++)
        {
            if (faces[face_a].neighbour_indices[edge_a] == -1)
            {
                bool edge_found = false;

                for (face_b = face_a + 1; face_b < faces.size(); face_b++)
                {
                    for (edge_b = 0; edge_b < 3; edge_b++)
                    {
                        if (faces[face_b].neighbour_indices[edge_b] == -1)
                        {
                            #define vertex_a1 vertices[faces[face_a].vertex_indices[edge_a]]
                            #define vertex_a2 vertices[faces[face_a].vertex_indices[(edge_a+1) % 3]]

                            #define vertex_b1 vertices[faces[face_b].vertex_indices[edge_b]]
                            #define vertex_b2 vertices[faces[face_b].vertex_indices[(edge_b+1) % 3]]

                            if (((vertex_a1 == vertex_b1) && (vertex_a2 == vertex_b2)) ||
                                ((vertex_a1 == vertex_b2) && (vertex_a2 == vertex_b1)))
                            {
                                faces[face_a].neighbour_indices[edge_a] = face_b;
                                faces[face_b].neighbour_indices[edge_b] = face_a;
                                edge_found = true;
                            }
                        }
                        if (edge_found) break;
                    }
                    if (edge_found) break;
                }
            }
        }
    }
/*
    std::cout << timer.GetElapsedTime() << std::endl;
*/
}

vector3 project_and_clip(const vector3& v, const vector3& light_vector_div_denom, const vector3 &plane_normal, const float& d)
{
    float numer = DotProduct(plane_normal, v) + d;

    return v - numer * light_vector_div_denom;
}

void Shadowable::cast_shadow(IGraphicsSystem* IGS, const matrix44& object_to_world_transform, Light current_light,
    ANIMATION_INSTANCE_HANDLE instance, shadow_mode mode)
{
    unsigned int f;
    vector3 light_vec = current_light.direction.normalize();
    vector3 light_vec_reverse = reverse_rotate(light_vec, object_to_world_transform).normalize();

    vector3 origin = transform(vector3(0.0f, 0.0f, 0.0f), object_to_world_transform);

    if (instance != NULL)
    {
        for (unsigned int v = 0; v < vertices.size(); v++)
        {
            vertices[v].x = instance[v * 6 + 0];
            vertices[v].y = instance[v * 6 + 1];
            vertices[v].z = instance[v * 6 + 2];

            // Normals are ignored here. instance[v * 6 + 3] is normal.x

        }
    }

    // Determine which faces are visible by the light
    for (f = 0; f < faces.size(); f++)
    {
        // Determine the normal of each face
        vector3 origin = vertices[faces[f].vertex_indices[0]];
        vector3 v1 = vertices[faces[f].vertex_indices[1]] - origin;
        vector3 v2 = vertices[faces[f].vertex_indices[2]] - origin;

        vector3 normal = CrossProduct(v1, v2); //.normalize();

        // Determine the angle between this and the light vector
        float dot = DotProduct(normal, light_vec_reverse);

        if (dot > 0.0f)
        {
            faces[f].visible = false;
        }
        else
        {
            faces[f].visible = true;
        }
    }

    // Light::DRECTIONAL
    float d;
    vector3 plane_normal, light_vec_reverse_div_denom;

    if (current_light.type == Light::DIRECTIONAL)
    {
        vector3 origin = reverse_transform(current_light.clip1, object_to_world_transform);
        vector3 plane2 = reverse_transform(current_light.clip2, object_to_world_transform);
        vector3 plane3 = reverse_transform(current_light.clip3, object_to_world_transform);

        plane_normal = CrossProduct(plane2 - origin, plane3 - origin).normalize();

        d = -DotProduct(plane_normal, origin);
        float denom = DotProduct(plane_normal, light_vec_reverse);
        light_vec_reverse_div_denom = (1.0f / denom) * light_vec_reverse;
    }


    std::vector<vector3> points;
    points.reserve(5000);

    if (mode == Z_FAIL_SHADOWS)
    {
        for (f = 0; f < faces.size(); f++)
        {
            if (faces[f].visible)
            {
                vector3 v[6];

                v[0] = vertices[faces[f].vertex_indices[0]];
                v[1] = vertices[faces[f].vertex_indices[1]];
                v[2] = vertices[faces[f].vertex_indices[2]];

                if (current_light.type == Light::DIRECTIONAL)
                {
                    v[3] = project_and_clip(v[0], light_vec_reverse_div_denom, plane_normal, d);
                    v[4] = project_and_clip(v[1], light_vec_reverse_div_denom, plane_normal, d);
                    v[5] = project_and_clip(v[2], light_vec_reverse_div_denom, plane_normal, d);
                }

                points.push_back(v[3]);
                points.push_back(v[5]);
                points.push_back(v[4]);

                points.push_back(v[0]);
                points.push_back(v[1]);
                points.push_back(v[2]);

                for (int edge = 0; edge < 3; edge++)
                {
                    int neighbour = faces[f].neighbour_indices[edge];

                    // If its neighbouring face is not visible, or there is no neighbour then this edge casts a shadow
                    if ((faces[neighbour].visible == false) || (neighbour == -1))
                    {
                        vector3 v1, v2, v3, v4;

                        // Get the points on the edge
                        v1 = v[edge];
                        v2 = v[(edge+1) % 3];

                        // Get the two vertices in the distance
                        v3 = v[edge + 3];
                        v4 = v[((edge+1) % 3) + 3];

                        points.push_back(v1);
                        points.push_back(v3);
                        points.push_back(v2);

                        points.push_back(v3);
                        points.push_back(v4);
                        points.push_back(v2);
                    }
                }
            }
        }
    }
    else // (mode == Z_PASS_SHADOWS)
    {
        for (f = 0; f < faces.size(); f++)
        {
            if (faces[f].visible)
            {
                for (int edge = 0; edge < 3; edge++)
                {
                    int neighbour = faces[f].neighbour_indices[edge];

                    // If its neighbouring face is not visible, or there is no neighbour then this edge casts a shadow
                    if ((faces[neighbour].visible == false) || (neighbour == -1))
                    {
                        vector3 v1, v2, v3, v4;

                        // Get the points on the edge
                        v1 = vertices[faces[f].vertex_indices[edge]];
                        v2 = vertices[faces[f].vertex_indices[(edge+1) % 3]];

                        if (current_light.type == Light::DIRECTIONAL)
                        {
                            // Get the two vertices in the distance
                            v3 = project_and_clip(v1, light_vec_reverse_div_denom, plane_normal, d);
                            v4 = project_and_clip(v2, light_vec_reverse_div_denom, plane_normal, d);
                        }

                        points.push_back(v1);
                        points.push_back(v3);
                        points.push_back(v2);

                        points.push_back(v3);
                        points.push_back(v4);
                        points.push_back(v2);
                    }
                }
            }
        }
    }

    IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, object_to_world_transform);

    IGS->StencilSetFirstShadowPass(mode);
    IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ, (float*)&points[0], points.size());

    IGS->StencilSetSecondShadowPass(mode);
    IGS->DrawTrianglesUP(TRIANGLE_LIST, VTX_XYZ, (float*)&points[0], points.size());
}

// Creates a Trimesh and adds it to the world.
dGeomID Shadowable::CreateTriMesh(dWorldID world, dSpaceID space, const vector3& position, float y_rotation)
{
    int number_of_vertices = vertices.size();
    int number_of_indices = indices.size();

    if (Vertices == NULL)
    {
        Vertices = new dVector3[number_of_vertices];
        Indices = new int[number_of_indices];

        for (int v = 0; v < number_of_vertices; v++)
        {
            Vertices[v][0] = vertices[v].x;
            Vertices[v][1] = vertices[v].y;
            Vertices[v][2] = vertices[v].z;
        }

        for (int i = 0; i < number_of_indices; i++)
        {
            Indices[i] = (int)indices[i];
        }

        TriMeshData = dGeomTriMeshDataCreate();

        dGeomTriMeshDataBuildSimple(TriMeshData, (dReal*)Vertices, number_of_vertices, Indices, number_of_indices);
    }

    dGeomID TriMesh = dCreateTriMesh(space, TriMeshData, 0, 0, 0);

    dGeomSetPosition(TriMesh, position.x, position.y, position.z);

    dQuaternion q;
    dQFromAxisAndAngle(q, 0.0, 1.0, 0.0, (dReal)y_rotation);
    dGeomSetQuaternion(TriMesh, q);

    return TriMesh;
}




