// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef SHADOWABLE_H
#define SHADOWABLE_H

#include "ode_thread.h"
#include "mtxlib.h"
#include "graphicssystem.h"

class Light;

//! A handle which references an animation instance
typedef float* ANIMATION_INSTANCE_HANDLE;

class Shadowable
{
public:
    //! Initialises the shadowable object.
    Shadowable();

    //! Deinitialises the shadowable object.
    virtual ~Shadowable();

    //! Returns the object geometry.
    virtual std::vector<vector3> GetVertices() = 0;

    //! Returns the indices for the object geometry.
    virtual std::vector<unsigned short> GetIndices() = 0;

    //! Draws this object on screen
    virtual void Render(ANIMATION_INSTANCE_HANDLE instance, Light current_light,
        matrix44 object_to_world_transform) = 0;

    //! Determines whether or not the object is in the view frustum. time is only applicable to animated models.
    virtual bool IsWithinFrustum(matrix44 object_to_world_transform, float time = 0.0) = 0;

    //! Determines whether or not camera is shadowed by this object.
    virtual bool IsCameraWithinShadow(matrix44 object_to_world_transform, vector3 camera, Light current_light, float time = 0.0) = 0;

    //! Determine if the shadow is within the view frustum.
    virtual bool IsShadowWithinFrustum(matrix44 object_to_world_transform, Light current_light, float time = 0.0) = 0;

    // Casts a shadow
    void cast_shadow(IGraphicsSystem* IGS, const matrix44& object_to_world_transform, Light current_light,
        ANIMATION_INSTANCE_HANDLE instance = NULL, shadow_mode mode = Z_PASS_SHADOWS);

    // Creates a Trimesh and adds it to the world.
    dGeomID CreateTriMesh(dWorldID world, dSpaceID space, const vector3& position, float y_rotation = 0.0);

protected:

    // Builds a connectivity list of all the faces in the object
    // Derived classes should call this at the end of setting up the object.
    void build_connectivity_list();

    std::vector<vector3> vertices;
    std::vector<unsigned short> indices;

private:

    struct face
    {
        int vertex_indices[3];      // Index of each vertex that forms this face
        int neighbour_indices[3];   // Index of each face that neighbours this one
        bool visible;               // Is the face visible to the light
    };

    std::vector<face> faces;

    // Stuff for the trimesh
    dTriMeshDataID TriMeshData;
    dVector3* Vertices;
    int* Indices;
};

#endif // SHADOWABLE_H

