// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "surface.h"
#include <stddef.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <cctype>
#include <algorithm>
// This extern "C" seems to fix some undefined references.
extern "C"
{
#include <jpeglib.h>
}
#include <png.h>

int GetNumberOfBytesPerPixel(surface_type type)
{
    if (type == R8G8B8)
    {
        return 3;
    }
    else if (type == A8R8G8B8 || type == R8G8B8A8)
    {
        return 4;
    }
    else if (type == X8)
    {
        return 1;
    }
    else
    {
        std::cout << "Invalid pixel format passed to GetNumberOfBytesPerPixel" << std::endl;
        return 0;
    }
}


Surface::Surface()
{
    width = 0;
    height = 0;
    buffer = NULL;
    type = INVALID;
}

Surface::~Surface()
{
    // Clear the surface.
    this->Clear();
}

Surface::Surface(int width, int height, surface_type type)
{
    this->width = width;
    this->height = height;
    this->type = type;

    this->buffer = new unsigned char[width * height * GetNumberOfBytesPerPixel(type)];
}

Surface::Surface(std::string filename)
{
    width = 0;
    height = 0;
    buffer = NULL;
    type = INVALID;

    this->Load(filename);
}

//! Copy constructor
Surface::Surface(const Surface& surface)
{
    width = surface.GetWidth();
    height = surface.GetHeight();
    buffer = NULL;
    type = surface.GetSurfaceType();

    this->buffer = new unsigned char[width * height * GetNumberOfBytesPerPixel(type)];

    memcpy(this->buffer, surface.buffer, width * height * GetNumberOfBytesPerPixel(type));
}

void Surface::operator=(const Surface& surface)
{
    width = surface.GetWidth();
    height = surface.GetHeight();
    buffer = NULL;
    type = surface.GetSurfaceType();

    this->buffer = new unsigned char[width * height * GetNumberOfBytesPerPixel(type)];

    memcpy(this->buffer, surface.buffer, width * height * GetNumberOfBytesPerPixel(type));
}

//! Deallocates memory used by the surface
void Surface::Clear()
{
    if (buffer != NULL)
    {
        delete[] buffer;
        buffer = NULL;
    }

    this->width  = 0;
    this->height = 0;
    this->type = INVALID;
}

//! Saves an image
void Surface::Save(std::string filename)
{
    // Convert the filename to lower case
    std::transform(filename.begin(), filename.end(), filename.begin(), (int(*)(int))tolower);

    if (filename.size() < 4)
    {
        return;
    }
    std::string last_four_chars = filename.substr(filename.size() - 4);

    if (last_four_chars == ".png")
    {
        this->SavePNG(filename);
    }
    else if (last_four_chars == "jpeg" || last_four_chars == ".jpg")
    {
        this->SaveJPEG(filename);
    }
    else
    {
        std::cout << "Filename passed to Surface::Save must end with .png or .jpg or .jpeg" << std::endl;
    }
}

//! Loads an image
void Surface::Load(std::string filename)
{
    // Convert the filename to lower case
    std::transform(filename.begin(), filename.end(), filename.begin(), (int(*)(int))tolower);

    if (filename.size() < 4)
    {
        return;
    }
    std::string last_four_chars = filename.substr(filename.size() - 4);

    if (last_four_chars == ".png")
    {
        this->LoadPNG(filename);
    }
    else if (last_four_chars == "jpeg" || last_four_chars == ".jpg")
    {
        this->LoadJPEG(filename);
    }
    else
    {
        std::cout << "Filename passed to Surface::Load must end with .png or .jpg or .jpeg" << std::endl;
    }
}

void Surface::LoadJPEG(std::string filename)
{
    // Clear the surface if it's already used.
    this->Clear();

    // 1. Allocate and initialize a JPEG decompression object.
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);

    // 2. Specify the source of the compressed data.
    FILE* infile;

    if ((infile = fopen(filename.c_str(), "rb")) == NULL)
    {
        std::stringstream error;
        error << "Can't open " << filename;
        throw std::string(error.str());
    }
    jpeg_stdio_src(&cinfo, infile);

    // 3. Call jpeg_read_header() to obtain image info.
    jpeg_read_header(&cinfo, TRUE);

    jpeg_calc_output_dimensions(&cinfo);

    this->width  = cinfo.output_width;
    this->height = cinfo.output_height;
    this->buffer = new unsigned char[width * height * 3];
    this->type = R8G8B8;

    // Decompress the image
    JSAMPROW rowptr[1];
    jpeg_start_decompress(&cinfo);
    while ( cinfo.output_scanline < cinfo.output_height )
    {
        rowptr[0] = (JSAMPROW)(unsigned char *)buffer +
            (cinfo.output_scanline * cinfo.output_width * 3);
        jpeg_read_scanlines(&cinfo, rowptr, (JDIMENSION) 1);
    }
    jpeg_finish_decompress(&cinfo);

    jpeg_destroy_decompress(&cinfo);

    fclose(infile);
}

void Surface::SaveJPEG(std::string filename, int quality)
{
    if (this->buffer == NULL)
    {
        throw std::string("this->buffer == NULL in function Surface::SaveJPEG");
    }
    if (this->type != R8G8B8)
    {
        throw std::string("this->type != R8G8B8 in function Surface::SaveJPEG");
    }

    // 1. Allocate and initialize a JPEG compression object.
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    // 2. Specify the destination for the compressed data (eg, a file).
    FILE* outfile;

    if ((outfile = fopen(filename.c_str(), "wb")) == NULL)
    {
        std::stringstream error;
        error << "Can't open " << filename;
        throw std::string(error.str());
    }
    jpeg_stdio_dest(&cinfo, outfile);

    // 3. Set parameters for compression, including image size & colorspace.
    cinfo.image_width = this->width;
    cinfo.image_height= this->height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;
    cinfo.optimize_coding = true;

    jpeg_set_defaults(&cinfo);

    jpeg_set_quality(&cinfo, quality, false);

    // 4. jpeg_start_compress(...);

    jpeg_start_compress(&cinfo, true);

    // 5. while (scan lines remain to be written)
    //     jpeg_write_scanlines(...);

    JSAMPROW row_pointer[1];    /* pointer to a single row */
    int row_stride;             /* physical row width in buffer */

    row_stride = this->width * 3;   /* JSAMPLEs per row in image_buffer */

    while (cinfo.next_scanline < cinfo.image_height)
    {
        row_pointer[0] = &this->buffer[cinfo.next_scanline * row_stride];
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }

    // 6. jpeg_finish_compress(...);
    jpeg_finish_compress(&cinfo);

    // 7. Release the JPEG compression object.
    jpeg_destroy_compress(&cinfo);

    fclose(outfile);
}

void Surface::LoadPNG(std::string filename)
{
    // Clear the surface if it's already used.
    this->Clear();

    FILE *fp = fopen(filename.c_str(), "rb");
    if (!fp)
    {
        std::stringstream error;
        error << "Can't open " << filename;
        throw std::string(error.str());
    }
    unsigned char header[4];
    int number = 4;
    fread(header, 1, number, fp);
    bool is_png = !png_sig_cmp(header, 0, number);
    if (!is_png)
    {
        std::stringstream error;
        error << filename << " is not a valid PNG file";
        throw std::string(error.str());
    }

    png_structp png_ptr = png_create_read_struct
       (PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
    if (!png_ptr)
    {
        std::stringstream error;
        error << "png_create_read_struct returned NULL";
        throw std::string(error.str());
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
        png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        std::stringstream error;
        error << "png_create_info_struct returned NULL";
        throw std::string(error.str());
    }

    png_infop end_info = png_create_info_struct(png_ptr);
    if (!end_info)
    {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        std::stringstream error;
        error << "png_create_info_struct returned NULL";
        throw std::string(error.str());
    }

    if (setjmp(png_jmpbuf(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr, &info_ptr,
           &end_info);
        std::stringstream error;
        error << "Error reading PNG file";
        throw std::string(error.str());
    }

    png_init_io(png_ptr, fp);

    png_set_sig_bytes(png_ptr, number);

    png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

    int bit_depth, color_type, interlace_type;

    png_get_IHDR(png_ptr, info_ptr, (png_uint_32*)&width, (png_uint_32*)&height,
       &bit_depth, &color_type, &interlace_type, NULL, NULL);

    if (info_ptr->channels == 1)
    {
        this->type = X8;
    }
    else if (info_ptr->channels == 3)
    {
        this->type = R8G8B8;
    }
    else if (info_ptr->channels == 4)
    {
        this->type = R8G8B8A8;
    }

    this->buffer = new unsigned char[width * height * info_ptr->channels];

    png_bytep* row_pointers;
    row_pointers = png_get_rows(png_ptr, info_ptr);

    for (int scanline = 0; scanline < this->height; scanline++)
    {
        memcpy(buffer + scanline * info_ptr->channels * width, row_pointers[scanline],
            info_ptr->channels * width);
    }

    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

    fclose(fp);
}

void Surface::SavePNG(std::string filename)
{
    if (this->buffer == NULL)
    {
        throw std::string("this->buffer == NULL in function Surface::SavePNG");
    }
    if (this->type != R8G8B8 && this->type != A8R8G8B8 && this->type != R8G8B8A8 && this->type != X8)
    {
        throw std::string("this->type is not a valid pixel format in function Surface::SavePNG");
    }

    FILE *fp = fopen(filename.c_str(), "wb");
    if (!fp)
    {
        std::stringstream error;
        error << "Can't open " << filename;
        throw std::string(error.str());
    }

    png_structp png_ptr = png_create_write_struct
       (PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
    if (!png_ptr)
    {
        throw std::string("Failed to png_create_write_struct");
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
        png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
        throw std::string("Failed to png_create_info_struct");
    }

    png_init_io(png_ptr, fp);

    // set the zlib compression level
    png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);

    if (this->type == X8)
    {
        png_set_IHDR(png_ptr, info_ptr, width, height,
            8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
            PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    }
    else if (this->type == R8G8B8)
    {
        png_set_IHDR(png_ptr, info_ptr, width, height,
            8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
            PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    }
    else // surface has alpha channel
    {
        png_set_IHDR(png_ptr, info_ptr, width, height,
            8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
            PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    }

    unsigned char** rows = new unsigned char*[height];
    for (int h = 0; h < height; h++)
    {
        rows[h] = this->buffer + (GetNumberOfBytesPerPixel(this->type) * width) * h;
    }

    png_set_rows(png_ptr, info_ptr, rows);

    if (this->type == A8R8G8B8)
    {
        png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_SWAP_ALPHA, NULL);
    }
    else // alpha does not need to be flipped
    {
        png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
    }

    png_destroy_write_struct(&png_ptr, &info_ptr);

    delete[] rows;

    fclose(fp);
}

//! Scales the surface down in size
void Surface::MinFilter(int dest_width, int dest_height)
{
    if (this->buffer == NULL)
    {
        throw std::string("this->buffer == NULL in function Surface::MinFilter");
    }
    if (this->type == INVALID)
    {
        throw std::string("this->type is not a valid pixel format in function Surface::MinFilter");
    }
    if (this->width % dest_width != 0)
    {
        throw std::string("this->width % dest_width != 0 in function Surface::MinFilter");
    }
    if (this->height % dest_height != 0)
    {
        throw std::string("this->height % dest_height != 0 in function Surface::MinFilter");
    }

    int components = GetNumberOfBytesPerPixel(this->type);

    unsigned char* new_buffer = new unsigned char[dest_width * dest_height * components];

    int stepx = this->width   / dest_width;
    int stepy = this->height  / dest_height;

    int source_pitch = this->width * components;
    int dest_pitch = dest_width * components;

    for (int j=0; j < dest_height; j++)
    {
        for (int i=0;i< dest_width; i++)
        {
            int component_totals[4];

            component_totals[0] = 0;
            component_totals[1] = 0;
            component_totals[2] = 0;
            component_totals[3] = 0;

            for (int y = 0; y < stepy ;y++)
            {
                for (int x = 0; x < stepx ;x++)
                {
                    for (int c = 0; c < components; c++)
                    {
                        unsigned char c_pixel = *(buffer + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * components + c);
                        component_totals[c] += c_pixel;
                    }
                }
            }

            int pixeltotal = (stepx * stepy);

            for (int c = 0; c < components; c++)
            {
                unsigned char c_pixel = component_totals[c] / pixeltotal;

                *(new_buffer + j * dest_pitch + i * components + c) = c_pixel;
            }
        }
    }

    delete[] this->buffer;
    this->buffer = new_buffer;
    this->width = dest_width;
    this->height= dest_height;
}

// Scales the surface up in size
void Surface::MagFilter(int dest_width, int dest_height)
{
    if (this->buffer == NULL)
    {
        throw std::string("this->buffer == NULL in function Surface::MagFilter");
    }
    if (this->type == INVALID)
    {
        throw std::string("this->type is not a valid pixel format in function Surface::MagFilter");
    }
    if (dest_width % this->width != 0)
    {
        throw std::string("dest_width % this->width != 0 in function Surface::MagFilter");
    }
    if (dest_height % this->height != 0)
    {
        throw std::string("dest_height % this->height != 0 in function Surface::MagFilter");
    }

    int components = GetNumberOfBytesPerPixel(this->type);

    unsigned char* new_buffer = new unsigned char[dest_width * dest_height * components];

    int ratio_x = dest_width / this->width;
    int ratio_y = dest_height / this->height;

    for (int j = 0; j < dest_height; j++)
    {
        for (int i = 0; i < dest_width; i++)
        {
            // This is a bilinear filter.
            int x = i / ratio_x;
            int y = j / ratio_y;

            for (int c = 0; c < components; c++)
            {
                unsigned char c1 = buffer[(   x    +    y    * this->width) * components + c];
                unsigned char c2 = buffer[((x + 1) +    y    * this->width) * components + c];
                unsigned char c3 = buffer[(   x    + (y + 1) * this->width) * components + c];
                unsigned char c4 = buffer[((x + 1) + (y + 1) * this->width) * components + c];

                // These two lines make less of an anomaly at the right and bottom edges.
                if (x == this->width - 1) { c2 = c1; c4 = c3; }
                if (y == this->height - 1) { c3 = c1; c4 = c2; }

                unsigned char c_top = c1 + (((c2 - c1) * (i%ratio_x))/ratio_x);
                unsigned char c_bot = c3 + (((c4 - c3) * (i%ratio_x))/ratio_x);

                unsigned char c_final = c_top + (((c_bot - c_top) * (j%ratio_y))/ratio_y);

                new_buffer[(i + j * dest_width) * components + c] = c_final;
            }
        }
    }

    delete[] this->buffer;
    this->buffer = new_buffer;
    this->width = dest_width;
    this->height= dest_height;
}

//! Returns a surface that is a rectangular part of another surface.
Surface Surface::GetSubSurface(int left, int top, int right, int bottom)
{
    int new_width = right - left;
    int new_height = bottom - top;
    int components = GetNumberOfBytesPerPixel(this->type);

    Surface result(new_width, new_height, type);

    for (int x = 0; x < new_width; x++)
    {
        for (int y = 0; y < new_height; y++)
        {
            for (int c = 0; c < components; c++)
            {
                result.buffer[(x + y * new_width) * components + c] =
                    this->buffer[(x + left + (y + top) * this->width) * components + c];
            }
        }
    }

    return result;
}

//! Rotates the surface. Degrees must be 90, 180, or 270. Surface should be square.
void Surface::Rotate(int degrees)
{
    int components = GetNumberOfBytesPerPixel(this->type);

    Surface temp(this->width, this->height, this->type);

    for (int x = 0; x < this->width; x++)
    {
        for (int y = 0; y < this->height; y++)
        {
            int x_new, y_new;

            if (degrees == 0)
            {
                x_new = x;
                y_new = y;
            }
            else if (degrees == 90)
            {
                x_new = y;
                y_new = this->width - x;
            }
            else if (degrees == 180)
            {
                x_new = this->width - x;
                y_new = this->height - y;
            }
            else // if (degrees == 270)
            {
                x_new = this->height - y;
                y_new = x;
            }

            for (int c = 0; c < components; c++)
            {
                temp.buffer[(x + y * this->width) * components + c] =
                    this->buffer[(x_new + y_new * this->width) * components + c];
            }
        }
    }

    memcpy(this->buffer, temp.buffer, this->width * this->height * components);
}

void Blend(unsigned char* s1, unsigned char* s2, unsigned char* dest, unsigned char amount)
{
/*
    for (int c = 0; c < GetNumberOfBytesPerPixel(type); c++)
    {
        dest[c] = (s1[c] * amount + s2[c] * (255 - amount)) / 255;
    }
*/
    dest[0] = (s1[0] * amount + s2[0] * (255 - amount)) / 255;
    dest[1] = (s1[1] * amount + s2[1] * (255 - amount)) / 255;
    dest[2] = (s1[2] * amount + s2[2] * (255 - amount)) / 255;
}

void Blend(unsigned char* s1, unsigned char r, unsigned char g, unsigned char b,
    unsigned char* dest, unsigned char amount)
{
    dest[0] = (s1[0] * amount + r * (255 - amount)) / 255;
    dest[1] = (s1[1] * amount + g * (255 - amount)) / 255;
    dest[2] = (s1[2] * amount + b * (255 - amount)) / 255;
}

void SetPixel(unsigned char* dest, unsigned char* src)
{
    dest[0] = src[0];
    dest[1] = src[1];
    dest[2] = src[2];
}



