// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef SURFACE_H
#define SURFACE_H

#include <string>

/*!
    \file surface.h
    \brief File containing Surface class for dealing with images.
*/

//! An enum of surface pixel formats.
enum surface_type
{
    R8G8B8,     //!< 24 bit.
    A8R8G8B8,   //!< 24 bit with 8bit alpha.
    R8G8B8A8,   //!< 24 bit with 8bit alpha.
    X8,         //!< 8 bit surface
    INVALID     //!< Invalid surface type.
};

//! Returns the number of bytes per pixel of a given pixel format.
/*!
    \param type pixel format to determine the number of bytes from.
    \return the number of bytes in the given pixel format.
*/
int GetNumberOfBytesPerPixel(surface_type type);

//! Surface class. Surface represents a rectangular image.
class Surface
{
public:
    //! Constructor
    Surface();

    //! Destructor
    ~Surface();

    //! Alternative constructor.
    /*!
        This version of the constructor allocates the memory for a new image given its
        width, height and pixel format.
        \param width the width in pixels of the new surface
        \param height the height in pixels of the new surface
        \param type the surface type of the new surface
    */
    Surface(int width, int height, surface_type type);

    //! Alternative constructor.
    /*!
        This version of the constructor loads a surface from file
        This function uses the filename extension to determine
        what file format to load. After calling this function the width, height and pixel data
        of this surface will be those of the loaded image.
        \param filename the name of the file to load.
    */
    Surface(std::string filename);

    //! Copy constructor
    Surface(const Surface& surface);

    //! Overloaded assignment operator.
    void operator=(const Surface& surface);

    //! Deallocates memory used by the surface
    /*!
        Deallocates memory used by the surface and sets the width and height to 0.
    */
    void Clear();

    //! Returns the width of the surface
    int GetWidth() const { return width; };

    //! Returns the height of the surface
    int GetHeight() const { return height; };

    //! Returns the pixel format of the surface
    surface_type GetSurfaceType() const { return type; };

    //! Saves an image
    /*!
        Saves this surface to an image file. This function uses the filename extension to determine
        what file format to save to.
        \param filename the filename to name the saved image.
    */
    void Save(std::string filename);

    //! Loads an image
    /*!
        Loads an image onto this surface. This function uses the filename extension to determine
        what file format to load. After calling this function the width, height and pixel data
        of this surface will be those of the loaded image.
        \param filename the name of the file to load.
    */
    void Load(std::string filename);

    //! Loads a jpeg.
    /*!
        This function loads a jpeg onto this surface. After calling this function the width,
        height and pixel data of this surface will be those of the jpeg.
        \param filename the name of the jpeg file to load.
    */
    void LoadJPEG(std::string filename);

    //! Saves a jpeg
    /*!
        Saves this surface to a jpeg. This surface must be R8G8B8 format.
        \param filename the filename to name the saved jpeg.
        \param quality the compression quality of the jpeg on a scale of 0..100
    */
    void SaveJPEG(std::string filename, int quality = 85);

    //! Loads a PNG.
    /*!
        This function loads a PNG onto this surface. After calling this function the width,
        height and pixel data of this surface will be those of the PNG.
        \param filename the name of the PNG file to load.
    */
    void LoadPNG(std::string filename);

    //! Saves a PNG
    /*!
        Saves this surface to a PNG file. This surface can be in R8G8B8, A8R8G8B8 or R8G8B8A8 format.
        \param filename the filename to name the saved PNG
    */
    void SavePNG(std::string filename);

    //! Scales the surface down in size
    /*!
        Uses a box filter to scale the surface down in size. After this the height and width will have changed
        and the buffer will also be reduced in size
        \param dest_width the width of the scaled down surface
        \param dest_height the height of the scaled down surface
    */
    void MinFilter(int dest_width, int dest_height);

    //! Scales the surface up in size
    /*!
        Uses a box filter to scale the surface up in size. After this the height and width will have changed
        and the buffer will also be bigger in size
        \param dest_width the width of the scaled up surface
        \param dest_height the height of the scaled up surface
    */
    void MagFilter(int dest_width, int dest_height);

    //! Returns a surface that is a rectangular part of another surface.
    Surface GetSubSurface(int left, int top, int right, int bottom);

    //! Used to determine whether the surface has an alpha channel
    bool HasAlphaChannel() { return (type == A8R8G8B8 || type == R8G8B8A8); }

    //! Array of pixel data
    unsigned char* buffer;

    //! Rotates the surface. Degrees must be 90, 180, or 270. Surface should be square.
    void Rotate(int degrees);

private:

    //! Width of surface in pixels.
    int width;

    //! Height of surface in pixels.
    int height;

    // Surface pixel format.
    surface_type type;
};

void Blend(unsigned char* s1, unsigned char* s2, unsigned char* dest, unsigned char amount);

void Blend(unsigned char* s1, unsigned char r, unsigned char g, unsigned char b,
    unsigned char* dest, unsigned char amount);

void SetPixel(unsigned char* dest, unsigned char* src);

#endif // SURFACE_H
