// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "transform.h"

vector3 transform(const vector3& v, const matrix44& mat)
{
    float x = v.x * mat[0][0] + v.y * mat[1][0] + v.z * mat[2][0] + mat[3][0];
    float y = v.x * mat[0][1] + v.y * mat[1][1] + v.z * mat[2][1] + mat[3][1];
    float z = v.x * mat[0][2] + v.y * mat[1][2] + v.z * mat[2][2] + mat[3][2];

    return vector3(x,y,z);
}

vector3 reverse_transform(const vector3& v, const matrix44& mat)
{
    float x = v.x - mat[3][0];
    float y = v.y - mat[3][1];
    float z = v.z - mat[3][2];

    return reverse_rotate(vector3(x,y,z), mat);
}

vector3 rotate(const vector3& v, const matrix44& mat)
{
    float x = v.x * mat[0][0] + v.y * mat[1][0] + v.z * mat[2][0];
    float y = v.x * mat[0][1] + v.y * mat[1][1] + v.z * mat[2][1];
    float z = v.x * mat[0][2] + v.y * mat[1][2] + v.z * mat[2][2];

    return vector3(x,y,z);
}

vector3 reverse_rotate(const vector3& v, const matrix44& mat)
{
    float x = v.x * mat[0][0] + v.y * mat[0][1] + v.z * mat[0][2];
    float y = v.x * mat[1][0] + v.y * mat[1][1] + v.z * mat[1][2];
    float z = v.x * mat[2][0] + v.y * mat[2][1] + v.z * mat[2][2];

    return vector3(x,y,z);
}

matrix44 concat_transforms(const matrix44& m1, const matrix44& m2)
{
    matrix44 result;

    result[0][0] = m1[0][0] * m2[0][0] + m1[1][0] * m2[0][1] + m1[2][0] * m2[0][2];
    result[1][0] = m1[0][0] * m2[1][0] + m1[1][0] * m2[1][1] + m1[2][0] * m2[1][2];
    result[2][0] = m1[0][0] * m2[2][0] + m1[1][0] * m2[2][1] + m1[2][0] * m2[2][2];
    result[3][0] = m1[0][0] * m2[3][0] + m1[1][0] * m2[3][1] + m1[2][0] * m2[3][2] + m1[3][0];
    result[0][1] = m1[0][1] * m2[0][0] + m1[1][1] * m2[0][1] + m1[2][1] * m2[0][2];
    result[1][1] = m1[0][1] * m2[1][0] + m1[1][1] * m2[1][1] + m1[2][1] * m2[1][2];
    result[2][1] = m1[0][1] * m2[2][0] + m1[1][1] * m2[2][1] + m1[2][1] * m2[2][2];
    result[3][1] = m1[0][1] * m2[3][0] + m1[1][1] * m2[3][1] + m1[2][1] * m2[3][2] + m1[3][1];
    result[0][2] = m1[0][2] * m2[0][0] + m1[1][2] * m2[0][1] + m1[2][2] * m2[0][2];
    result[1][2] = m1[0][2] * m2[1][0] + m1[1][2] * m2[1][1] + m1[2][2] * m2[1][2];
    result[2][2] = m1[0][2] * m2[2][0] + m1[1][2] * m2[2][1] + m1[2][2] * m2[2][2];
    result[3][2] = m1[0][2] * m2[3][0] + m1[1][2] * m2[3][1] + m1[2][2] * m2[3][2] + m1[3][2];

    return result;
}

matrix44 angle_matrix(const vector3& angles)
{
    matrix44 result;

    float angle;
    float sr, sp, sy, cr, cp, cy;

    angle = DegToRad(angles[2]);
    sy = sin(angle);
    cy = cos(angle);
    angle = DegToRad(angles[1]);
    sp = sin(angle);
    cp = cos(angle);
    angle = DegToRad(angles[0]);
    sr = sin(angle);
    cr = cos(angle);

    // result = (Z * Y) * X
    result[0][0] = cp*cy;
    result[0][1] = cp*sy;
    result[0][2] = -sp;
    result[1][0] = sr*sp*cy+cr*-sy;
    result[1][1] = sr*sp*sy+cr*cy;
    result[1][2] = sr*cp;
    result[2][0] = (cr*sp*cy+-sr*-sy);
    result[2][1] = (cr*sp*sy+-sr*cy);
    result[2][2] = cr*cp;
    result[3][0] = 0.0;
    result[3][1] = 0.0;
    result[3][2] = 0.0;

    return result;
}

void angle_quaternion(const vector3& angles, quaternion& q)
{
    float angle;
    float sr, sp, sy, cr, cp, cy;

    angle = angles[2] * 0.5;
    sy = sin(angle);
    cy = cos(angle);
    angle = angles[1] * 0.5;
    sp = sin(angle);
    cp = cos(angle);
    angle = angles[0] * 0.5;
    sr = sin(angle);
    cr = cos(angle);

    q[0] = sr*cp*cy-cr*sp*sy; // X
    q[1] = cr*sp*cy+sr*cp*sy; // Y
    q[2] = cr*cp*sy-sr*sp*cy; // Z
    q[3] = cr*cp*cy+sr*sp*sy; // W
}


matrix44 quaternion_matrix(const quaternion& q)
{
    matrix44 matrix;

    matrix[0][0] = 1.0 - 2.0 * q[1] * q[1] - 2.0 * q[2] * q[2];
    matrix[0][1] = 2.0 * q[0] * q[1] + 2.0 * q[3] * q[2];
    matrix[0][2] = 2.0 * q[0] * q[2] - 2.0 * q[3] * q[1];

    matrix[1][0] = 2.0 * q[0] * q[1] - 2.0 * q[3] * q[2];
    matrix[1][1] = 1.0 - 2.0 * q[0] * q[0] - 2.0 * q[2] * q[2];
    matrix[1][2] = 2.0 * q[1] * q[2] + 2.0 * q[3] * q[0];

    matrix[2][0] = 2.0 * q[0] * q[2] + 2.0 * q[3] * q[1];
    matrix[2][1] = 2.0 * q[1] * q[2] - 2.0 * q[3] * q[0];
    matrix[2][2] = 1.0 - 2.0 * q[0] * q[0] - 2.0 * q[1] * q[1];

    return matrix;
}

void quaternion_slerp(quaternion& p, quaternion& q, float t, quaternion& qt)
{
    int i;
    float omega, cosom, sinom, sclp, sclq;

    // decide if one of the quaternions is backwards
    float a = 0;
    float b = 0;
    for (i = 0; i < 4; i++)
    {
        a += (p[i]-q[i])*(p[i]-q[i]);
        b += (p[i]+q[i])*(p[i]+q[i]);
    }
    if (a > b)
    {
        for (i = 0; i < 4; i++)
        {
            q[i] = -q[i];
        }
    }

    cosom = p[0]*q[0] + p[1]*q[1] + p[2]*q[2] + p[3]*q[3];

    if ((1.0 + cosom) > 0.00000001)
    {
        if ((1.0 - cosom) > 0.00000001)
        {
            omega = acos( cosom );
            sinom = sin( omega );
            sclp = sin( (1.0 - t)*omega) / sinom;
            sclq = sin( t*omega ) / sinom;
        }
        else
        {
            sclp = 1.0 - t;
            sclq = t;
        }
        for (i = 0; i < 4; i++)
        {
            qt[i] = sclp * p[i] + sclq * q[i];
        }
    }
    else
    {
        qt[0] = -p[1];
        qt[1] = p[0];
        qt[2] = -p[3];
        qt[3] = p[2];
        sclp = sin( (1.0 - t) * 0.5 * 3.1428);
        sclq = sin( t * 0.5 * 3.1428);
        for (i = 0; i < 3; i++)
        {
            qt[i] = sclp * p[i] + sclq * qt[i];
        }
    }
}








