// ============================================================================
// These functions are taken from Telemachos' PXDTUT10.
// ============================================================================
//
// Desc: A collection of vector math related functions.
//
// Copyright (c) 2000 Telemachos of Peroxide
// www.peroxide.dk

#include "vector_math.h"
#include <iostream>

// ----------------------------------------------------------------------
// Name  : closestPointOnLine()
// Input : a - first end of line segment
//         b - second end of line segment
//         p - point we wish to find closest point on line from
// Notes : Helper function for closestPointOnTriangle()
// Return: closest point on line segment
// -----------------------------------------------------------------------
vector3 closestPointOnLine(const vector3& a, const vector3& b, const vector3& p)
{
    // Determine t (the length of the vector from a to p)
    vector3 c = p-a;
    vector3 V = b-a;

    double d = V.length();

    V.normalize();
    double t = DotProduct(V,c);

    // Check to see if t is beyond the extents of the line segment
    if (t < 0.0f) return (a);
    if (t > d) return (b);

    // Return the point between a and b
    //set length of V to t. V is normalized so this is easy
    V.x = V.x * t;
    V.y = V.y * t;
    V.z = V.z * t;

    return (a + V);
}

// ----------------------------------------------------------------------
// Name  : closestPointOnTriangle()
// Input : a - first vertex in triangle
//         b - second vertex in triangle
//         c - third vertex in triangle
//         p - point we wish to find closest point on triangle from
// Notes :
// Return: closest point on line triangle edge
// -----------------------------------------------------------------------

vector3 closestPointOnTriangle(const vector3& a, const vector3& b, const vector3& c, const vector3& p)
{
    vector3 Rab = closestPointOnLine(a, b, p);
    vector3 Rbc = closestPointOnLine(b, c, p);
    vector3 Rca = closestPointOnLine(c, a, p);

    double dAB = (vector3(p-Rab)).length();
    double dBC = (vector3(p-Rbc)).length();
    double dCA = (vector3(p-Rca)).length();

    double min = dAB;
    vector3 result = Rab;

    if (dBC < min)
    {
        min = dBC;
        result = Rbc;
    }

    if (dCA < min)
    {
        result = Rca;
    }

    return (result);
}

// ----------------------------------------------------------------------
// Name  : intersectRayPlane()
// Input : rOrigin - origin of ray in world space
//         rVector - vector describing direction of ray in world space
//         pOrigin - Origin of plane
//         pNormal - Normal to plane
// Notes : Normalized directional vectors expected
// Return: distance to plane in world units, -1 if no intersection.
// -----------------------------------------------------------------------
float intersectRayPlane(const vector3 &rOrigin, const vector3 &rVector, const vector3 &pOrigin, const vector3 &pNormal)
{
    float d = - (DotProduct(pNormal,pOrigin));

    float numer = DotProduct(pNormal,rOrigin) + d;
    float denom = DotProduct(pNormal,rVector);

    if (denom == 0)  // normal is orthogonal to vector, can't intersect
        return (-1.0f);

    return -(numer / denom);
}

// ----------------------------------------------------------------------
// Name  : CheckPointInTriangle()
// Input : point - point we wish to check for inclusion
//         a - first vertex in triangle
//         b - second vertex in triangle
//         c - third vertex in triangle
// Notes : Triangle should be defined in clockwise order a,b,c
// Return: TRUE if point is in triangle, FALSE if not.
// -----------------------------------------------------------------------

bool CheckPointInTriangle(const vector3 &point, const vector3 &a, const vector3 &b, const vector3 &c)
{
    double total_angles = 0.0;

    // make the 3 vectors
    vector3 v1 = point-a;
    vector3 v2 = point-b;
    vector3 v3 = point-c;

    v1.normalize();
    v2.normalize();
    v3.normalize();

    total_angles += acos(DotProduct(v1,v2));
    total_angles += acos(DotProduct(v2,v3));
    total_angles += acos(DotProduct(v3,v1));

    if (fabs(total_angles-(2.0f*3.1428f)) <= 0.005f)
        return (true);

    return(false);
}

// ----------------------------------------------------------------------
// Name  : closestPointOnTriangleSurface()
// Input : a - first vertex in triangle
//         b - second vertex in triangle
//         c - third vertex in triangle
//         p - point we wish to find closest point on triangle from
//         closest - resulting point. The closest point to p on the triangles surface
// Notes :
// Return: true if closest point is valid. false if invalid.
// -----------------------------------------------------------------------
bool closestPointOnTriangleSurface(const vector3 &a, const vector3 &b, const vector3 &c,
                                        const vector3 &p, vector3 &closest, float radius)
{
    // Make the plane containing this triangle.
    vector3 pOrigin = a;
    vector3 v1 = b-a;
    vector3 v2 = c-a;

    // determine normal to plane containing polygon
    vector3 pNormal = CrossProduct(v1, v2).normalize();

    // find how far away the plane is from point p along the planes normal
    float distToPlaneIntersection = intersectRayPlane(p, -pNormal, pOrigin, pNormal);
    if ((distToPlaneIntersection == -1) || (distToPlaneIntersection > radius)) return false;

    // find the nearest point on the plane to p
    closest = p - (pNormal * distToPlaneIntersection);

    // determine if that point is in the triangle
    return CheckPointInTriangle(closest, a, b, c);
}

/*
// Return: true if ray intersects triangle
bool intersectRayTriangle(const vector3 &a, const vector3 &b, const vector3 &c,
                            const vector3 &rOrigin, const vector3 &rVector, vector3& contact)
{
    // Make the plane containing this triangle.
    vector3 pOrigin = a;
    vector3 v1 = b-a;
    vector3 v2 = c-a;

    // determine normal to plane containing polygon
    vector3 pNormal = CrossProduct(v1, v2).normalize();

    // find how far away the plane is from point rOrigin along the ray
    float distToPlaneIntersection = intersectRayPlane(rOrigin, rVector, pOrigin, pNormal);
    if ((distToPlaneIntersection == -1)) return false;

    if ((distToPlaneIntersection < 0)) return false;

    // find the nearest point on the plane to rOrigin
    contact = rOrigin + rVector * distToPlaneIntersection;

    // determine if that point is in the triangle
    return CheckPointInTriangle(contact, a, b, c);
}
*/

