// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "world.h"
#include "modelfactory.h"
#include "object_renderer.h"
#include "filenames.h"
#include <fstream>
#include <sstream>
#include <iostream>

World::Object::Object(std::string filename, IGraphicsSystem* IGS)
{
    this->model_filename = filename;

    this->model = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
    this->model->Load(data_dir() + slash() + "world" + slash(), model_filename, IGS);
}

World::Location::Location(IModel* model, vector3 v, float rotation_y)
{
    this->position = v;
    this->rotation_y = rotation_y;

    dWorldID world = OdeThread::Instance().getWorld();
    dSpaceID space = OdeThread::Instance().getSpace();

    this->model_geom = model->CreateTriMesh(world, space, position, rotation_y);
}

World::World(IGraphicsSystem* IGS)
{
    this->IGS = IGS;
/*
    Object main("level001.txt", IGS);

    Location main_location(main.model);
    main.instances.push_back(main_location);

    objects.push_back(main);

    Object underpass("underpass001.txt", IGS);

    Location underpass1(underpass.model, vector3(-128.0 - 64.0, 0.0, -512.0 - 256.0),DegToRad(90.0));
    underpass.instances.push_back(underpass1);

    Location underpass2(underpass.model, vector3(-128.0 - 64.0 + 512.0, 0.0, -512.0 - 256.0), DegToRad(90.0));
    underpass.instances.push_back(underpass2);

    objects.push_back(underpass);
*/
}

World::~World()
{
    // Clear the world
    New();
}

void World::Render()
{
    // Render each instance of all the models.
    for (unsigned int o = 0; o < objects.size(); o++)
    {
        for (unsigned int i = 0; i < objects[o].instances.size(); i++)
        {
            vector3 pos = objects[o].instances[i].position;
            float rot_y = objects[o].instances[i].rotation_y;

            matrix44 transform = TranslateMatrix44(pos.x, pos.y, pos.z) * RotateRadMatrix44('Y', rot_y);

            ObjectRenderer::Instance().AddModel(objects[o].model, transform);
        }
    }
}

void World::New()
{
    // Delete all the models and the model geoms
    for (unsigned int o = 0; o < objects.size(); o++)
    {
        delete objects[o].model;

        for (unsigned int i = 0; i < objects[o].instances.size(); i++)
        {
            dGeomDestroy(objects[o].instances[i].model_geom);
        }

        objects[o].instances.clear();
    }

    objects.clear();
}

void World::Load(std::string filename)
{
    New();

    std::ifstream input_file;
    input_file.open(filename.c_str());
    if (input_file.fail())
    {
        std::stringstream error;
        error << "Failed to open " << filename;
        throw std::string(error.str());
    }

    std::string s;
    int number_of_objects;

    input_file >> s;                // Ignore "Objects:"
    input_file >> number_of_objects;

    for (int o = 0; o < number_of_objects; o++)
    {
        std::string model_filename;
        input_file >> model_filename;
        Object obj(model_filename, IGS);

        int number_of_instances;
        input_file >> s;            // Ignore "Instances:"
        input_file >> number_of_instances;

        for (int i = 0; i < number_of_instances; i++)
        {
            vector3 pos;
            float rot_y;

            input_file >> pos.x >> pos.y >> pos.z >> rot_y;

            Location loc(obj.model, pos, rot_y);

            obj.instances.push_back(loc);
        }

        objects.push_back(obj);
    }

    input_file.close();
}

void World::Save(std::string filename)
{
    std::ofstream output_file;
    output_file.open(filename.c_str());
    if (output_file.fail())
    {
        std::stringstream error;
        error << "Failed to open " << filename;
        throw std::string(error.str());
    }

    output_file << "Objects: " << objects.size() << std::endl;

    for (unsigned int o = 0; o < objects.size(); o++)
    {
        output_file << objects[o].model_filename << std::endl;

        output_file << "Instances: " << objects[o].instances.size() << std::endl;

        for (unsigned int i = 0; i < objects[o].instances.size(); i++)
        {
            vector3 pos = objects[o].instances[i].position;
            float rot_y = objects[o].instances[i].rotation_y;

            output_file << pos.x << ' ' << pos.y << ' ' << pos.z << ' ' << rot_y << std::endl;
        }
    }

    output_file.close();
}

void World::AddInstance(std::string filename, vector3 v, float rotation_y)
{
    // Check if there is an object of the same type already in the world
    for (unsigned int o = 0; o < objects.size(); o++)
    {
        if (objects[o].model_filename == filename)
        {
            // This object is of the same type

            // Check if there is an instance identical to this one
            for (unsigned int i = 0; i < objects[o].instances.size(); i++)
            {
                if (objects[o].instances[i].position == v && objects[o].instances[i].rotation_y == rotation_y)
                {
                    return;
                }
            }

            // No existing instance is identical to this one. Add a new instance.
            Location loc(objects[o].model, v, rotation_y);
            objects[o].instances.push_back(loc);
            return;
        }
    }

    // There is no object of this type already in the world. Add a new object.
    Object obj(filename, IGS);

    Location loc(obj.model, v, rotation_y);
    obj.instances.push_back(loc);

    objects.push_back(obj);
}



